#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause


"""Builder of CST segments."""

import logging
import os
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union

from cryptography.hazmat.primitives.ciphers.aead import AESCCM
from cryptography.hazmat.primitives.serialization import Encoding

from spsdk.crypto import Certificate
from spsdk.crypto.loaders import load_certificate_from_data, load_private_key_from_data
from spsdk.exceptions import (
    SPSDKAttributeError,
    SPSDKError,
    SPSDKFileNotFoundError,
    SPSDKKeyError,
    SPSDKTypeError,
    SPSDKValueError,
)
from spsdk.image import segments
from spsdk.image.commands import (
    CmdAuthData,
    CmdBase,
    CmdInstallKey,
    CmdSet,
    CmdUnlockAbstract,
    CmdUnlockCAAM,
    CmdUnlockSNVS,
    EnumAuthDat,
    EnumCertFormat,
    EnumEngine,
    EnumInsKey,
)
from spsdk.image.hab.config_parser import ImageConfig, SectionConfig
from spsdk.image.hab.hab_binary_image import HabBinaryImage, HabSegment
from spsdk.image.header import Header, SegTag
from spsdk.image.images import BootImgRT
from spsdk.image.secret import MAC, CertificateImg, EnumAlgorithm, Signature, SrkTable
from spsdk.utils.crypto.common import crypto_backend
from spsdk.utils.misc import (
    BinaryPattern,
    align_block,
    find_file,
    find_first,
    get_abs_path,
    load_binary,
    write_file,
)

logger = logging.getLogger(__name__)


@dataclass
class ImageBlock:
    """Single image block."""

    base_address: int
    start: int
    size: int


class SecCommand(ABC):
    """Sec command abstract class."""

    CMD_INDEX: int
    CONFIGURATION_PARAMS: Dict[str, Any]

    def __init__(self) -> None:
        """Command abstract class constructor."""
        self._cmd: Optional[Union[CmdBase, Header]] = None

    @property
    def cmd(self) -> Union[CmdBase, Header]:
        """Command property.

        :raises SPSDKAttributeError: If command is not set
        """
        if self._cmd is None:
            raise SPSDKAttributeError("Command is not set")
        return self._cmd

    @cmd.setter
    def cmd(self, value: Union[CmdBase, Header]) -> None:
        """Command setter."""
        self._cmd = value

    @classmethod
    def check_config_section_params(cls, section_data: SectionConfig) -> None:
        """Check if options contains only accepted arguments.

        :param section_data: Section data to be checked
        :raises SPSDKError: If mandatory parameter is not present
        :raises SPSDKError: If unexpected key is present
        """
        for param, is_mandatory in cls.CONFIGURATION_PARAMS.items():
            if is_mandatory and section_data.options.get(param) is None:
                raise SPSDKError("Mandatory parameter is not defined")

        additional_params = list(
            set(key.lower() for key in section_data.options.keys())
            - set(key.lower() for key in cls.CONFIGURATION_PARAMS.keys())
        )
        if additional_params:
            raise SPSDKError(f"Unexpected arguments {additional_params}")

    @staticmethod
    @abstractmethod
    def parse(config: SectionConfig, search_paths: Optional[List[str]] = None) -> "SecCommand":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Parsed command instance
        """

    @abstractmethod
    def build_command(self) -> None:
        """Build command with given properties."""

    @staticmethod
    def generate_random_bytes(length: int) -> bytes:
        """Generate random bytes.

        :param length: Length of random bytes
        :raises SPSDKError: If length of bytes is not as expected
        :return: Generated random bytes
        """
        secret_key = crypto_backend().random_bytes(length)
        if length != len(secret_key):
            raise SPSDKError(f"Invalid sectet key bytes length: {len(secret_key)}")
        return secret_key


class SecCsfHeader(SecCommand):
    """CSF header command."""

    CMD_INDEX = 20
    CONFIGURATION_PARAMS = {
        "Header_Version": True,
        "Header_HashAlgorithm": False,
        "Header_Engine": True,
        "Header_EngineConfiguration": True,
        "Header_CertificateFormat": False,
        "Header_SignatureFormat": True,
    }

    def __init__(
        self,
        version: Union[int, str],
        engine: EnumEngine,
        engine_config: int,
        hash_algorithm: Optional[EnumAlgorithm] = None,
        certificate_format: Optional[EnumCertFormat] = None,
        signature_format: Optional[str] = None,
    ) -> None:
        """CSF header class constructor.

        :param version: Header version
        :param engine: Engine plugin tag
        :param engine_config: Engine configuration index
        :param hash_algorithm: Hash algorithm type
        :param certificate_format: Certificate format tag
        :param signature_format: Signature format
        :raises SPSDKValueError: Invalid combination of input parameters.
        :raises SPSDKValueError: Invalid signature format.
        """
        super().__init__()
        self.version = self._parse_version(version)
        if engine == EnumEngine.ANY and engine_config != 0:
            raise SPSDKValueError(f"Invalid combination:{engine}: {engine_config}")
        self.engine: EnumEngine = engine
        self.engine_config: int = engine_config
        if not signature_format:
            signature_format = "CMS"
        if signature_format and signature_format.lower() != "cms":
            raise SPSDKValueError(f"Invalid signature format: {signature_format}")
        self.signature_format: str = signature_format
        self.certificate_format: Optional[EnumCertFormat] = certificate_format
        self.hash_algorithm: Optional[EnumAlgorithm] = hash_algorithm

    @staticmethod
    def parse(config: SectionConfig, search_paths: Optional[List[str]] = None) -> "SecCsfHeader":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfHeader.check_config_section_params(config)
        version = config.options["Header_Version"]
        assert version

        hash_algorithm = config.options.get("Header_HashAlgorithm")
        if hash_algorithm is not None:
            hash_algorithm = EnumAlgorithm[hash_algorithm]

        engine = EnumEngine[config.options["Header_Engine"]]
        engine_config = int(config.options["Header_EngineConfiguration"])

        cert_format = config.options.get("Header_CertificateFormat")
        if cert_format is not None:
            cert_format = EnumCertFormat[cert_format]

        signature_format = config.options.get("Header_SignatureFormat")
        return SecCsfHeader(
            version=version,
            engine=engine,
            engine_config=engine_config,
            hash_algorithm=hash_algorithm,
            certificate_format=cert_format,
            signature_format=signature_format,
        )

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: Header = Header(SegTag.CSF, param=self.version)
        self.cmd.length = 4

    @staticmethod
    def _parse_version(version: Union[str, int]) -> int:
        """Parse version from string to actiual integer.

        An example: "4.2" -> 0x42 -> 64

        :param version: Version as string or int
        :raises SPSDKTypeError: Input version if wrong type.
        :return: Version as int value
        """
        if not isinstance(version, (str, int)):
            raise SPSDKTypeError("Version can be either int or string")
        if isinstance(version, str):
            int_version = int(version.replace(".", ""), 16)
        else:
            int_version = version
        return int_version


class SecCsfInstallSrk(SecCommand):
    """Install SRK command."""

    CMD_INDEX = 21
    CONFIGURATION_PARAMS = {"InstallSRK_Table": True, "InstallSRK_SourceIndex": True}

    def __init__(self, srk_table: bytes, source_index: int) -> None:
        """Install SRK class constructor.

        :param srk_table: SRK table
        :param source_index: Source index
        :raises SPSDKError: Srk table is not defined .
        :raises SPSDKError: Source index is not defined .
        """
        super().__init__()
        if not srk_table:
            raise SPSDKError("SRK table must be defined.")
        self.srk_table = srk_table
        if source_index is None:
            raise SPSDKError("Source index must be defined.")
        self.source_index = source_index

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecCsfInstallSrk":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfInstallSrk.check_config_section_params(config)

        source_index = int(config.options["InstallSRK_SourceIndex"])
        srk_table = load_binary(config.options["InstallSRK_Table"], search_paths=search_paths)

        return SecCsfInstallSrk(srk_table=srk_table, source_index=source_index)

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd = CmdInstallKey(flags=EnumInsKey.CLR, hash_alg=EnumAlgorithm.SHA256, tgt_index=0)

        self.cmd.certificate_ref = SrkTable.parse(self.srk_table)
        self.cmd.source_index = self.source_index


class SecCsfInstallCsfk(SecCommand):
    """Install CSFK command."""

    CMD_INDEX = 22
    CONFIGURATION_PARAMS = {"InstallCSFK_File": True, "InstallCSFK_CertificateFormat": False}

    def __init__(
        self,
        csfk_file_path: str,
        certificate_format: Optional[EnumCertFormat] = None,
    ) -> None:
        """Install CSF class constructor.

        :param csfk_file_path: Path to CSFK file
        :param version: Header version
        :param certificate_format: Certificate format
        :raises SPSDKValueError: Invalid combination of input parameters.
        """
        super().__init__()
        self._version: Optional[int] = None
        self.csfk_file_path = csfk_file_path
        if certificate_format == EnumCertFormat.SRK:
            raise SPSDKValueError(f"Invalid certificate format: {EnumCertFormat.SRK}")
        if certificate_format is None:
            certificate_format = EnumCertFormat.X509
        self.certificate_format = certificate_format

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecCsfInstallCsfk":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfInstallCsfk.check_config_section_params(config)

        csfk_file_path = find_file(config.options["InstallCSFK_File"], search_paths=search_paths)
        cert_format = config.options.get("InstallCSFK_CertificateFormat")
        cert_format = EnumCertFormat[cert_format] if cert_format is not None else None

        return SecCsfInstallCsfk(
            csfk_file_path=csfk_file_path,
            certificate_format=cert_format,
        )

    @property
    def version(self) -> int:
        """Image version."""
        assert self._version is not None
        return self._version

    @version.setter
    def version(self, value: int) -> None:
        """Image version setter."""
        if not isinstance(value, int):
            raise SPSDKTypeError("Version must be int type")
        self._version = value

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdInstallKey = CmdInstallKey(
            flags=EnumInsKey.CSF,
            tgt_index=1,
        )
        self.cmd.certificate_format = self.certificate_format  # type: ignore
        certificate_bin = load_binary(self.csfk_file_path)
        certificate = load_certificate_from_data(certificate_bin)
        self.cmd.certificate_ref = CertificateImg(
            version=self.version, data=certificate.public_bytes(Encoding.DER)
        )


class SecCsfAuthenticateCsf(SecCommand):
    """Authenticate CSFK command."""

    CMD_INDEX = 24
    CONFIGURATION_PARAMS = {
        "AuthenticateCsf_PrivateKeyFile": False,
        "AuthenticateCsf_KeyPass": False,
    }
    SIGNED_DATA_SIZE = 768

    def __init__(
        self,
        private_key: Optional[bytes],
        key_pass: Optional[str] = None,
    ) -> None:
        """Authenticate CSFK class constructor.

        :param version: Header version
        :param certificate: Certificate
        :param private_key: Private key used for authentication
        :param data: Command data to be signed
        :param key_pass: Key for decryption of private key
        """
        super().__init__()
        self.private_key = private_key
        self.key_pass = key_pass
        self._version: Optional[int] = None
        self._certificate: Optional[Certificate] = None
        self._engine: Optional[EnumEngine] = None

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecCsfAuthenticateCsf":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfAuthenticateCsf.check_config_section_params(config)
        private_key_path = config.options.get("AuthenticateCsf_PrivateKeyFile")
        private_key = None
        if private_key_path is not None:
            private_key = load_binary(private_key_path, search_paths=search_paths)

        key_pass = None
        private_key_pass_path = config.options.get("AuthenticateCsf_KeyPass")
        if private_key_pass_path is not None:
            private_key_pass_file = find_file(private_key_pass_path, search_paths=search_paths)
            with open(private_key_pass_file) as file:
                key_pass = file.readline().strip()
        return SecCsfAuthenticateCsf(private_key=private_key, key_pass=key_pass)

    @property
    def certificate(self) -> Certificate:
        """Certificate."""
        assert self._certificate is not None
        return self._certificate

    @certificate.setter
    def certificate(self, value: Certificate) -> None:
        """Certificate setter."""
        if not isinstance(value, Certificate):
            raise SPSDKTypeError("Certificate is of incorrect type")
        self._certificate = value

    @property
    def engine(self) -> Optional[EnumEngine]:
        """Engine."""
        return self._engine

    @engine.setter
    def engine(self, value: Optional[EnumEngine]) -> None:
        """Engine setter."""
        if not (isinstance(value, EnumEngine) or isinstance(value, int)):
            raise SPSDKTypeError("Engine is of incorrect type")
        self._engine = value

    @property
    def version(self) -> int:
        """Image version."""
        assert self._version is not None
        return self._version

    @version.setter
    def version(self, value: int) -> None:
        """Image version setter."""
        if not isinstance(value, int):
            raise SPSDKTypeError("Version must be int type")
        self._version = value

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdAuthData = CmdAuthData(certificate=self.certificate)
        if self.engine is not None:
            self.cmd.engine = self.engine

        assert self.private_key is not None
        pem_private_key = load_private_key_from_data(
            self.private_key, password=str.encode(self.key_pass) if self.key_pass else None
        )
        self.cmd.private_key_pem_data = pem_private_key  # type: ignore

        signature = Signature(version=self.version)
        self.cmd.signature = signature

    def sign(self, data: bytes, timestamp: datetime) -> None:
        """Sign data and update command signature.

        :param data: Data to be signed
        :param timestamp: Signature timestamp
        """
        data = align_block(data, alignment=self.SIGNED_DATA_SIZE, padding=BinaryPattern("zeros"))
        self.cmd.update_signature(zulu=timestamp, data=data)


class SecCsfInstallKey(SecCommand):
    """Install key command."""

    CMD_INDEX = 25
    CONFIGURATION_PARAMS = {
        "InstallKey_File": True,
        "InstallKey_VerificationIndex": True,
        "InstallKey_TargetIndex": True,
    }

    def __init__(
        self,
        certificate_path: str,
        source_index: int,
        target_index: int,
    ) -> None:
        """Install key class constructor.

        :param certificate_path: Path to certificate
        :param version: Header version
        :param source_index: Source index
        :param target_index: Target index
        """
        super().__init__()
        self.certificate_path = certificate_path
        self.certificate = load_certificate_from_data(load_binary(certificate_path))
        self.source_index = source_index
        self.target_index = target_index
        self._version: Optional[int] = None

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecCsfInstallKey":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfInstallKey.check_config_section_params(config)

        install_key_path = config.options["InstallKey_File"]
        certificate_path = find_file(install_key_path, search_paths=search_paths)

        source_index = int(config.options["InstallKey_VerificationIndex"])

        target_index = int(config.options["InstallKey_TargetIndex"])
        return SecCsfInstallKey(
            certificate_path=certificate_path, source_index=source_index, target_index=target_index
        )

    @property
    def version(self) -> int:
        """Image version."""
        assert self._version is not None
        return self._version

    @version.setter
    def version(self, value: int) -> None:
        """Image version setter."""
        if not isinstance(value, int):
            raise SPSDKTypeError("Version must be int type")
        self._version = value

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdInstallKey = CmdInstallKey(cert_fmt=EnumCertFormat.X509)
        self.cmd.certificate_ref = CertificateImg(
            version=self.version, data=self.certificate.public_bytes(Encoding.DER)
        )
        self.cmd.source_index = self.source_index
        self.cmd.target_index = self.target_index


class SecCsfAuthenticateData(SecCommand):
    """Authenticate data command."""

    CMD_INDEX = 26
    KEY_IDX_AUT_DAT_FAST_AUTH = 0
    KEY_IDX_AUT_DAT_MIN = 2
    KEY_IDX_AUT_DAT_MAX = 5
    CONFIGURATION_PARAMS = {
        "AuthenticateData_VerificationIndex": True,
        "AuthenticateData_Engine": True,
        "AuthenticateData_EngineConfiguration": True,
        "AuthenticateData_PrivateKeyFile": False,
        "AuthenticateData_KeyPass": False,
    }

    def __init__(
        self,
        engine: EnumEngine,
        engine_config: int,
        verification_index: int,
        private_key: Optional[bytes],
        key_pass: Optional[str] = None,
    ) -> None:
        """Authenticate data class constructor.

        :param engine: Engine plugin tag
        :param engine_config: Engine configuration index
        :param verification_index: Target index
        :raises SPSDKValueError: Invalid combination of input parameters.
        :raises SPSDKValueError: Verification index is not defined.
        :raises SPSDKValueError: Key index is not a valid value.
        """
        super().__init__()
        if engine == EnumEngine.ANY and engine_config != 0:
            raise SPSDKValueError(f"Invalid argument combination:{engine}: {engine_config}")
        self.engine = engine
        self.engine_config = engine_config
        if verification_index is None:
            raise SPSDKValueError("Verification index must be defined.")
        if verification_index != self.KEY_IDX_AUT_DAT_FAST_AUTH and (
            verification_index < self.KEY_IDX_AUT_DAT_MIN
            or verification_index > self.KEY_IDX_AUT_DAT_MAX
        ):
            raise SPSDKValueError("Key index must have valid value.")
        self.verification_index = verification_index
        self.private_key = private_key
        self.key_pass = key_pass
        self._certificate: Optional[Certificate] = None
        self._blocks: List[ImageBlock] = []

    @property
    def blocks(self) -> List[ImageBlock]:
        """Blocks to be signed property."""
        return self._blocks

    @blocks.setter
    def blocks(self, value: List[ImageBlock]) -> None:
        """Blocks to be signed property setter."""
        if not isinstance(value, List):
            raise SPSDKTypeError("Blocks must be a list type")
        self._blocks = value

    def sign(
        self,
        data: bytes,
        base_data_address: int,
        timestamp: Optional[datetime] = None,
    ) -> None:
        """Sign data and update command signature.

        :param data: Data to be signed
        :param base_data_address: Base address of the generated data
        :param timestamp: Signature timestamp
        """
        if timestamp is None:
            timestamp = datetime.now(timezone.utc)
        self.cmd.update_signature(
            zulu=timestamp,
            data=data,
            base_data_addr=base_data_address,
        )

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecCsfAuthenticateData":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecCsfAuthenticateData.check_config_section_params(config)
        verification_index = int(config.options["AuthenticateData_VerificationIndex"])
        engine = EnumEngine[config.options["AuthenticateData_Engine"]]
        engine_cfg = int(config.options["AuthenticateData_EngineConfiguration"])

        private_key_path = config.options.get("AuthenticateData_PrivateKeyFile")
        private_key = None
        if private_key_path is not None:
            private_key = load_binary(private_key_path, search_paths=search_paths)

        key_pass = None
        private_key_pass_path = config.options.get("AuthenticateData_KeyPass")
        if private_key_pass_path is not None:
            private_key_pass_file = find_file(private_key_pass_path, search_paths=search_paths)
            with open(private_key_pass_file) as file:
                key_pass = file.readline().strip()
        return SecCsfAuthenticateData(
            verification_index=verification_index,
            engine=engine,
            engine_config=engine_cfg,
            private_key=private_key,
            key_pass=key_pass,
        )

    @property
    def certificate(self) -> Certificate:
        """Certificate."""
        assert self._certificate is not None
        return self._certificate

    @certificate.setter
    def certificate(self, value: Certificate) -> None:
        """Certificate setter."""
        if not isinstance(value, Certificate):
            raise SPSDKTypeError("Certificate is of incorrect type")
        self._certificate = value

    @property
    def version(self) -> int:
        """Image version."""
        assert self._version is not None
        return self._version

    @version.setter
    def version(self, value: int) -> None:
        """Image version setter."""
        if not isinstance(value, int):
            raise SPSDKTypeError("Version must be int type")
        self._version = value

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdAuthData = CmdAuthData(
            certificate=self.certificate,
        )
        self.cmd.engine = self.engine
        self.cmd.engine_cfg = self.engine_config
        self.cmd.key_index = self.verification_index

        assert self.private_key is not None
        pem_private_key = load_private_key_from_data(
            self.private_key, password=str.encode(self.key_pass) if self.key_pass else None
        )
        self.cmd.private_key_pem_data = pem_private_key  # type: ignore

        signature = Signature(version=self.version)
        self.cmd.signature = signature

        if self.blocks:
            for block in self.blocks:
                self.cmd.append(block.base_address, block.size)

    @staticmethod
    def _get_image_blocks(
        image_config: ImageConfig,
        app_length: int,
        is_encrypted: bool,
        search_paths: Optional[List[str]] = None,
    ) -> List[ImageBlock]:
        """Get image blocks from image binary.

        :param image_config: Loaded image configuration
        :param app_length: App length
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        blocks = []
        # IVT + BDT
        blocks.append(
            ImageBlock(
                base_address=image_config.options.start_address + image_config.options.ivt_offset,
                start=image_config.options.ivt_offset,
                size=segments.SegIVT2.SIZE + BootImgRT.BDT_SIZE,
            )
        )
        if image_config.options.dcd_file_path is not None:
            dcd_bin = load_binary(image_config.options.dcd_file_path, search_paths=search_paths)
            blocks.append(
                ImageBlock(
                    base_address=image_config.options.start_address
                    + image_config.options.ivt_offset
                    + segments.SegIVT2.SIZE
                    + BootImgRT.BDT_SIZE,
                    start=image_config.options.ivt_offset
                    + segments.SegIVT2.SIZE
                    + BootImgRT.BDT_SIZE,
                    size=len(dcd_bin),
                )
            )
        if image_config.options.xmcd_file_path is not None:
            xmcd_bin = load_binary(image_config.options.xmcd_file_path, search_paths=search_paths)

            blocks.append(
                ImageBlock(
                    base_address=image_config.options.start_address
                    + image_config.options.ivt_offset
                    + BootImgRT.XMCD_IVT_OFFSET,
                    start=image_config.options.ivt_offset + BootImgRT.XMCD_IVT_OFFSET,
                    size=len(xmcd_bin),
                )
            )
        if not is_encrypted:
            blocks.append(
                ImageBlock(
                    base_address=image_config.options.start_address
                    + image_config.options.initial_load_size,
                    start=image_config.options.initial_load_size,
                    size=app_length,
                )
            )
        return blocks


class SecSetEngine(SecCommand):
    """Set engine command."""

    CMD_INDEX = 31
    CONFIGURATION_PARAMS = {
        "SetEngine_HashAlgorithm": False,
        "SetEngine_Engine": False,
        "SetEngine_EngineConfiguration": False,
    }

    def __init__(
        self,
        hash_algorithm: Optional[EnumAlgorithm] = None,
        engine: Optional[EnumEngine] = None,
        engine_cfg: Optional[int] = None,
    ) -> None:
        """Set engine class constructor.

        :param hash_algorithm: Hash algorithm type
        :param engine: Engine plugin tag
        :param engine_config: Engine configuration index
        """
        super().__init__()
        self.hash_algorithm = hash_algorithm
        self.engine = engine
        self.engine_cfg = engine_cfg

    @staticmethod
    def parse(config: SectionConfig, search_paths: Optional[List[str]] = None) -> "SecSetEngine":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecSetEngine.check_config_section_params(config)
        hash_algorithm = config.options.get("SetEngine_HashAlgorithm")
        hash_algorithm = EnumAlgorithm[hash_algorithm] if hash_algorithm is not None else None

        engine = config.options.get("SetEngine_Engine")
        engine = EnumEngine[engine] if engine is not None else None

        engine_cfg = config.options.get("SetEngine_EngineConfiguration")
        engine_cfg = int(engine_cfg) if engine_cfg is not None else None
        return SecSetEngine(hash_algorithm=hash_algorithm, engine=engine, engine_cfg=engine_cfg)

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd = CmdSet()
        if self.hash_algorithm is not None:
            self.cmd.hash_algorithm = self.hash_algorithm

        if self.engine is not None:
            self.cmd.engine = self.engine

        if self.engine_cfg is not None:
            self.cmd.engine_cfg = self.engine_cfg


class SecUnlock(SecCommand):
    """Unlock engine command."""

    CMD_INDEX = 33
    ENGINE_CLASSES = {"SNVS": CmdUnlockSNVS, "CAAM": CmdUnlockCAAM}
    UNLOCK_FEARTURES = {"LP SWR": 1, "ZMK WRITE": 2}
    CONFIGURATION_PARAMS = {"Unlock_Engine": True, "Unlock_Features": False}

    def __init__(self, unlock_engine: str, features: Optional[int]) -> None:
        """Unlock class constructor.

        :param unlock_engine: Unlock engine type: can be aither SNVS or CAAM
        :param features: Features
        :raises SPSDKKeyError: Unknown engine.
        """
        super().__init__()
        if unlock_engine not in self.ENGINE_CLASSES:
            raise SPSDKKeyError(f"Unknown engine {unlock_engine}")
        self.unlock_engine = unlock_engine
        self.features = features

    @staticmethod
    def parse(config: SectionConfig, search_paths: Optional[List[str]] = None) -> SecCommand:
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        :raises SPSDKKeyError: Unknown features.
        """
        SecUnlock.check_config_section_params(config)
        unlock_engine = config.options["Unlock_Engine"]
        unlock_features = config.options.get("Unlock_Features")
        if unlock_features is not None:
            if unlock_features not in SecUnlock.UNLOCK_FEARTURES:
                raise SPSDKKeyError(f"Unknown features {unlock_features}")
            unlock_features = SecUnlock.UNLOCK_FEARTURES[unlock_features]
        return SecUnlock(unlock_engine=unlock_engine, features=unlock_features)

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdUnlockAbstract = self.ENGINE_CLASSES[self.unlock_engine]()
        if self.features is not None:
            self.cmd.features = self.features


class SecInstallSecretKey(SecCommand):
    """Set engine command."""

    CMD_INDEX = 27
    CONFIGURATION_PARAMS = {
        "SecretKey_Name": True,
        "SecretKey_Length": False,
        "SecretKey_VerifyIndex": False,
        "SecretKey_TargetIndex": True,
        "SecretKey_ReuseDek": False,
    }

    def __init__(
        self,
        secret_key: bytes,
        source_index: int,
        target_index: int,
    ) -> None:
        """Set install secret key class constructor.

        :param hash_algorithm: Hash algorithm type
        :param engine: Engine plugin tag
        :param engine_config: Engine configuration index
        :raises SPSDKValueError: Source index not specified.
        :raises SPSDKValueError: Source index is not lower or equal to 3.
        :raises SPSDKValueError: Target index not specified.
        """
        super().__init__()
        self.secret_key: bytes = secret_key
        if source_index is None:
            raise SPSDKValueError("Source index must be specified")
        if source_index > 3:
            raise SPSDKValueError("Source index must be equal or lower than 3")
        self.source_index = source_index
        if target_index is None:
            raise SPSDKValueError("Target index must be specified")
        self.target_index = target_index
        self._location: Optional[int] = None

    @staticmethod
    def parse(
        config: SectionConfig, search_paths: Optional[List[str]] = None
    ) -> "SecInstallSecretKey":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecInstallSecretKey.check_config_section_params(config)
        reuse_dek = True if config.options.get("SecretKey_ReuseDek", 0) == 1 else False
        length = int(config.options.get("SecretKey_Length", 128))  # type: ignore
        if length not in [128, 192, 256]:
            raise SPSDKValueError(f"Invalid sectet key length {length}")
        key_length = length // 8
        if reuse_dek:
            secret_key_path = find_file(config.options["SecretKey_Name"], search_paths=search_paths)
            secret_key = load_binary(secret_key_path)
        else:
            base_dir = search_paths[0] if search_paths is not None else None
            secret_key_path = get_abs_path(config.options["SecretKey_Name"], base_dir)
            secret_key = SecInstallSecretKey.generate_random_bytes(length // 8)
            SecInstallSecretKey.save_secret_key(secret_key_path, secret_key)
        if len(secret_key) != key_length:
            raise SPSDKError(
                f"Loaded secret key lenght does not match the expected length: {length}"
            )
        source_index = int(config.options.get("SecretKey_VerifyIndex", 0))  # type: ignore
        target_index = int(config.options["SecretKey_TargetIndex"])
        return SecInstallSecretKey(
            secret_key=secret_key, source_index=source_index, target_index=target_index
        )

    @staticmethod
    def save_secret_key(secret_key_path: str, secret_key: bytes) -> None:
        """Save given sectret key into file.

        :param secret_key_path: Path to file with secret key
        :param secret_key: Secret key to be saved
        """
        exists = os.path.isfile(secret_key_path)
        if exists:
            os.remove(secret_key_path)
        write_file(secret_key, secret_key_path, "wb")

    @property
    def location(self) -> int:
        """Start address of DEK key."""
        assert self._location is not None
        return self._location

    @location.setter
    def location(self, value: int) -> None:
        """Setter for start address of DEK key."""
        if not isinstance(value, int):
            raise SPSDKTypeError("Location must be int type")
        self._location = value

    def build_command(self) -> None:
        """Build command with given properties."""
        assert self.location is not None
        self.cmd: CmdInstallKey = CmdInstallKey(
            flags=EnumInsKey.ABS,
            cert_fmt=EnumCertFormat.BLOB,
            hash_alg=EnumAlgorithm.ANY,
            location=self.location,
        )
        self.cmd.source_index = self.source_index
        self.cmd.target_index = self.target_index


class SecDecryptData(SecCommand):
    """Set engine command."""

    CMD_INDEX = 28
    CONFIGURATION_PARAMS = {
        "Decrypt_Engine": False,
        "Decrypt_EngineConfiguration": False,
        "Decrypt_VerifyIndex": True,
        "Decrypt_MacBytes": False,
        "Decrypt_Nonce": False,
    }

    def __init__(
        self,
        verification_index: int,
        mac_len: Optional[int],
        nonce: Optional[bytes] = None,
        engine: Optional[EnumEngine] = None,
        engine_config: Optional[int] = None,
    ) -> None:
        """Decrypt data class constructor.

        :param verification_index: Target index
        :param mac_len: Number of mac bytes
        :param nonce: Nonce binary
        :param engine: Engine plugin tag
        :param engine_config: Engine configuration index
        :raises SPSDKValueError: Invalid combination of input parameters.
        """
        super().__init__()
        if verification_index is None:
            raise SPSDKValueError("Verification index must be defined.")
        if verification_index >= 6:
            raise SPSDKValueError("Verification index must be lower than 6.")
        self.verification_index = verification_index
        self.engine = engine if engine is not None else EnumEngine.ANY
        self.engine_config = engine_config if engine_config is not None else 0
        if engine == EnumEngine.ANY and engine_config != 0:
            raise SPSDKValueError(f"Invalid argument combination:{engine}: {engine_config}")
        self.mac_len = mac_len if mac_len is not None else 16
        if self.mac_len < 4 or self.mac_len > 16 or self.mac_len % 2:
            raise SPSDKValueError(
                "Invalid mac length. Valid options are 4, 6, 8, 10, 12, 14 and 16."
            )
        self.nonce = nonce
        self._dek: Optional[bytes] = None
        self._blocks: Optional[List[ImageBlock]] = None

    def encrypt(self, data: bytes) -> Tuple[bytes, bytes]:
        """Encrypt data and return mac and encrypted data.

        :raises SPSDKError: Invalid length of encrypted data.
        """
        assert self.dek is not None
        data_to_encrypt: bytes = b""
        if self.blocks:
            for block in self.blocks:
                data_to_encrypt += data[block.start : block.start + block.size]
        else:
            data_to_encrypt = data
        if self.nonce is None:
            nonce_len = BootImgRT.aead_nonce_len(len(data_to_encrypt))
            self.nonce = SecDecryptData.generate_random_bytes(nonce_len)
        aesccm = AESCCM(self.dek, tag_length=self.mac_len)
        encr = aesccm.encrypt(self.nonce, data_to_encrypt, None)  # type: ignore
        if len(encr) != len(data_to_encrypt) + self.mac_len:
            raise SPSDKError("Invalid length of encrypted data")
        mac = encr[-self.mac_len :]
        data = encr[: -self.mac_len]
        return mac, data

    @property
    def dek(self) -> bytes:
        """Data encryption key."""
        assert self._dek is not None
        return self._dek

    @dek.setter
    def dek(self, value: bytes) -> None:
        """Data encryption key setter."""
        if not isinstance(value, bytes):
            raise SPSDKTypeError("Dek must be bytes type")
        self._dek = value

    @property
    def blocks(self) -> Optional[List[ImageBlock]]:
        """Blocks to be encrypted property."""
        return self._blocks

    @blocks.setter
    def blocks(self, value: List[ImageBlock]) -> None:
        """Blocks to be encrypted property setter."""
        if not isinstance(value, List):
            raise SPSDKTypeError("Blocks must be a list type")
        self._blocks = value

    @staticmethod
    def parse(config: SectionConfig, search_paths: Optional[List[str]] = None) -> "SecDecryptData":
        """Parse configuration into the command.

        :param config: Section config
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        SecDecryptData.check_config_section_params(config)
        engine = config.options.get("Decrypt_Engine")
        if engine is not None:
            engine = EnumEngine[engine]

        engine_cfg = config.options.get("Decrypt_EngineConfiguration")
        if engine_cfg is not None:
            engine_cfg = int(engine_cfg)

        verification_index = int(config.options["Decrypt_VerifyIndex"])

        mac_len = config.options.get("Decrypt_MacBytes")
        if mac_len is not None:
            mac_len = int(mac_len)

        nonce = config.options.get("Decrypt_Nonce")
        if nonce is not None:
            nonce = load_binary(nonce, search_paths=search_paths)
        return SecDecryptData(
            verification_index=verification_index,
            mac_len=mac_len,
            nonce=nonce,
            engine=engine,
            engine_config=engine_cfg,
        )

    def build_command(self) -> None:
        """Build command with given properties."""
        self.cmd: CmdAuthData = CmdAuthData(flags=EnumAuthDat.CLR, sig_format=EnumCertFormat.AEAD)

        self.cmd.engine = self.engine
        self.cmd.engine_cfg = self.engine_config
        self.cmd.key_index = self.verification_index
        if self.blocks:
            for block in self.blocks:
                self.cmd.append(block.base_address, block.size)

    @staticmethod
    def _get_image_blocks(
        image_config: ImageConfig,
        app_length: int,
    ) -> List[ImageBlock]:
        blocks = []
        blocks.append(
            ImageBlock(
                base_address=image_config.options.start_address
                + image_config.options.initial_load_size,
                start=image_config.options.initial_load_size,
                size=app_length,
            )
        )
        return blocks


class CsfBuilder:
    """Csf command builder."""

    def __init__(
        self,
        bd_config: ImageConfig,
        csf_offset: int,
        hab_image: HabBinaryImage,
        search_paths: Optional[List[str]] = None,
        timestamp: Optional[datetime] = None,
    ) -> None:
        """CSF builder class constructor.

        :param bd_config: Loaded image configuration
        :param csf_offset: CSF segment offset
        :param hab_image: Hab binary image
        :param search_paths: List of paths where to search for the file, defaults to None
        :param timestamp: Signature timestamp
        """
        self.header: Optional[SecCsfHeader] = None
        self.commands: List[SecCommand] = []
        self.bd_config = bd_config
        self.is_encrypted = self.bd_config.options.flags == 0x0C
        self.is_authenticated = self.bd_config.options.flags == 0x08
        self.csf_offset = csf_offset
        self.search_paths = search_paths
        self.timestamp = timestamp if timestamp is not None else datetime.now(timezone.utc)
        if self.is_authenticated or self.is_encrypted:
            hab_image.align_segment(HabSegment.APP, 16)
        self.hab_image = hab_image

    def reset(self) -> None:
        """Reset builder into its initial state."""
        self.header = None
        self.commands = []

    def get_command(self, command_id: int, raise_exc: bool = True) -> Optional[SecCommand]:
        """Get command by command id.

        :param command_id: Command ID to be retrieved
        :param raise_exc: If set and section is not found, the error is raised
        :raises SPSDKKeyError: If command does not exist
        """
        command = find_first(self.commands, lambda cmd: cmd.CMD_INDEX == command_id)
        if raise_exc and command is None:
            raise SPSDKKeyError(f"Section with id {command_id} does not exist.")
        return command

    def append_command(self, command: SecCommand) -> None:
        """Append command to list of commands and update header length.

        :param command: Command to be appended
        """
        assert self.header
        assert self.header.cmd
        header = self.header.cmd
        header.length += len(command.cmd.export())
        self.commands.append(command)

    def get_padding_hab_image(self) -> HabBinaryImage:
        """Get HAB image with initial padding."""
        image = deepcopy(self.hab_image)
        for sub_img in image.sub_images:
            sub_img.offset += self.bd_config.options.ivt_offset
        return image

    @property
    def keyblob_address(self) -> int:
        """Keyblob address property."""
        return (
            self.bd_config.options.start_address
            + self.csf_offset
            + self.bd_config.options.ivt_offset
            + self.hab_image.CSF_SIZE
        )

    def build_csf_header(self) -> None:
        """Build CSF header command."""
        bd_section = self.bd_config.get_section(SecCsfHeader.CMD_INDEX)
        if not bd_section:
            return
        command = SecCsfHeader.parse(bd_section, search_paths=self.search_paths)
        command.build_command()
        self.header = command

    def build_csf_install_srk(self) -> None:
        """Build CSF install SRK command."""
        bd_section = self.bd_config.get_section(SecCsfInstallSrk.CMD_INDEX)
        if not bd_section:
            return
        command = SecCsfInstallSrk.parse(bd_section, search_paths=self.search_paths)
        command.build_command()
        self.append_command(command)

    def build_csf_install_csfk(self) -> None:
        """Build CSF install CSFK command."""
        bd_section = self.bd_config.get_section(SecCsfInstallCsfk.CMD_INDEX)
        if not bd_section:
            return

        assert self.header is not None
        command = SecCsfInstallCsfk.parse(bd_section, search_paths=self.search_paths)
        command.version = self.header.version
        command.build_command()

        self.append_command(command)

    def build_authenticate_csfk(self) -> None:
        """Build authenticate CSFK command."""
        bd_section = self.bd_config.get_section(SecCsfAuthenticateCsf.CMD_INDEX)
        if not bd_section:
            return

        assert self.header is not None
        install_csfk: SecCsfInstallCsfk = self.get_command(SecCsfInstallCsfk.CMD_INDEX)  # type: ignore

        command = SecCsfAuthenticateCsf.parse(bd_section, search_paths=self.search_paths)
        if command.private_key is None:
            private_key_path = self._determine_private_key_path(install_csfk.csfk_file_path)
            if not private_key_path:
                raise SPSDKFileNotFoundError("Private key could not be found.")
            private_key = load_binary(private_key_path, search_paths=self.search_paths)
            command.private_key = private_key

        if command.key_pass is None:
            key_pass_file = self._determine_key_pass_path(install_csfk.csfk_file_path)
            if key_pass_file:
                with open(key_pass_file) as file:
                    command.key_pass = file.readline().strip()

        command.certificate = load_certificate_from_data(
            load_binary(install_csfk.csfk_file_path, search_paths=self.search_paths)
        )
        command.version = self.header.version
        command.engine = self.header.engine
        command.build_command()

        # This is just temporary signature.
        # It will be recreated in the finish method
        data_to_sign = self.header.cmd.export()
        for section in self.commands:
            data_to_sign += section.cmd.export()
        data_to_sign += command.cmd.export()

        command.sign(data_to_sign, self.timestamp)
        self.append_command(command)

    def build_install_key_csfk(self) -> None:
        """Build install key CSFK command."""
        bd_section = self.bd_config.get_section(SecCsfInstallKey.CMD_INDEX)
        if not bd_section:
            return

        assert self.header is not None
        command = SecCsfInstallKey.parse(bd_section, search_paths=self.search_paths)
        command.version = self.header.version
        command.build_command()
        self.append_command(command)

    def build_authenticate_data(self) -> None:
        """Build authenticate data command."""
        bd_section = self.bd_config.get_section(SecCsfAuthenticateData.CMD_INDEX)
        if not bd_section:
            return

        assert self.header is not None
        install_key: SecCsfInstallKey = self.get_command(SecCsfInstallKey.CMD_INDEX)  # type: ignore

        command = SecCsfAuthenticateData.parse(bd_section, search_paths=self.search_paths)
        if command.private_key is None:
            private_key_path = self._determine_private_key_path(install_key.certificate_path)
            if not private_key_path:
                raise SPSDKFileNotFoundError("Private key could not be found.")
            command.private_key = load_binary(private_key_path, search_paths=self.search_paths)

        if command.key_pass is None:
            key_pass_file = self._determine_key_pass_path(install_key.certificate_path)
            if key_pass_file:
                with open(key_pass_file) as file:
                    command.key_pass = file.readline().strip()

        app = self.hab_image.get_hab_segment(HabSegment.APP)
        blocks = SecCsfAuthenticateData._get_image_blocks(
            self.bd_config,
            app_length=len(app),
            is_encrypted=self.is_encrypted,
            search_paths=self.search_paths,
        )

        command.certificate = install_key.certificate
        command.version = self.header.version
        command.blocks = blocks
        command.build_command()

        padding_image = self.get_padding_hab_image()
        command.sign(
            data=padding_image.export(),
            base_data_address=self.bd_config.options.start_address,
            timestamp=self.timestamp,
        )
        self.append_command(command)

    def build_set_engine(self) -> None:
        """Build set engine command."""
        bd_section = self.bd_config.get_section(SecSetEngine.CMD_INDEX)
        if not bd_section:
            return
        command = SecSetEngine.parse(bd_section, search_paths=self.search_paths)
        command.build_command()
        self.append_command(command)

    def build_unlock_engine(self) -> None:
        """Build unlock engine command."""
        bd_section = self.bd_config.get_section(SecUnlock.CMD_INDEX)
        if not bd_section:
            return
        command = SecUnlock.parse(bd_section, search_paths=self.search_paths)
        command.build_command()
        self.append_command(command)

    def build_install_secret_key(self) -> None:
        """Build install Secret key command.

        :raises SPSDKError: Incorrect version is used
        """
        bd_section = self.bd_config.get_section(SecInstallSecretKey.CMD_INDEX)
        if not bd_section:
            return

        assert self.header is not None

        if self.header.version <= 0x40:
            raise SPSDKError("The command is supported from version 0x41 onwards")

        command = SecInstallSecretKey.parse(bd_section, search_paths=self.search_paths)
        command.location = self.keyblob_address
        command.build_command()
        self.append_command(command)

    def build_decrypt_data(self) -> None:
        """Build install Secret key command."""
        bd_section = self.bd_config.get_section(SecDecryptData.CMD_INDEX)
        if not bd_section:
            return

        install_secret_key: SecInstallSecretKey = self.get_command(SecInstallSecretKey.CMD_INDEX)  # type: ignore

        command = SecDecryptData.parse(bd_section, search_paths=self.search_paths)
        command.dek = install_secret_key.secret_key
        app = self.hab_image.get_hab_segment(HabSegment.APP)
        blocks = SecDecryptData._get_image_blocks(
            self.bd_config,
            app_length=app.aligned_length(16),
        )
        command.blocks = blocks
        command.build_command()
        self.append_command(command)

    def finish(self) -> None:
        """Finish command creation."""
        if not self.commands:
            return None
        assert self.header is not None
        # Encrypt data if needed
        if self.is_encrypted:
            encrypt_data_cmd: SecDecryptData = self.get_command(SecDecryptData.CMD_INDEX)  # type: ignore
            app = self.hab_image.get_hab_segment(HabSegment.APP)
            assert app is not None

            padding_image = self.get_padding_hab_image()
            mac, data = encrypt_data_cmd.encrypt(padding_image.export())

            app.binary = data
            # Update signature
            decrypt: SecDecryptData = self.get_command(SecDecryptData.CMD_INDEX)  # type: ignore
            assert decrypt.nonce is not None
            decrypt.cmd.signature = MAC(
                version=self.header.version,
                nonce_len=len(decrypt.nonce),
                mac_len=decrypt.mac_len,
                data=decrypt.nonce + mac,
            )

        auth_command: SecCsfAuthenticateCsf = self.get_command(SecCsfAuthenticateCsf.CMD_INDEX)  # type: ignore
        # Sign all command's data at the end of whole process
        if auth_command is not None:
            csf = self._build_csf_segment()
            data_to_sign = self.header.cmd.export()
            for section in csf.commands:
                data_to_sign += section.export()
            auth_command.cmd.update_signature(zulu=self.timestamp, data=data_to_sign)
        self.hab_image.add_hab_segment(
            HabSegment.CSF, self._build_csf_segment().export(), offset_override=self.csf_offset
        )

    def _build_csf_segment(self) -> segments.SegCSF:
        """Build whole CSF segment from individual commands."""
        assert self.header is not None
        csf = segments.SegCSF(version=self.header.version, enabled=True)
        for section in self.commands:
            csf.append_command(section.cmd)  # type: ignore
        csf.update(True)
        return csf

    def _determine_private_key_path(self, cert_file_path: str) -> Optional[str]:
        """Determine private key path the same was as legacy CST tool does.

        :param cert_file_path: Path to certificate file
        :return: Path to private key file
        """
        logger.debug("Trying to determine the private key path.")
        keys_dir = self._get_keys_dir(cert_file_path)
        cert_file_name = os.path.basename(cert_file_path)
        cert_file, cert_extension = os.path.splitext(cert_file_name)
        if cert_file.endswith("crt"):
            key_file = cert_file[:-3] + "key"
            key_file_name = key_file + cert_extension
        else:
            return None
        private_key_file = os.path.join(keys_dir, key_file_name)
        if not os.path.isfile(private_key_file):
            return None
        return private_key_file

    def _determine_key_pass_path(self, cert_file_path: str) -> Optional[str]:
        """Determine key pass path the same was as legacy CST tool does.

        :param cert_file_path: Path to certificate file
        :return: Path to key pass file
        """
        logger.debug("Trying to determine the key pass path.")
        keys_dir = self._get_keys_dir(cert_file_path)
        key_pass_file = os.path.join(keys_dir, "key_pass.txt")
        if not os.path.isfile(key_pass_file):
            return None
        return key_pass_file

    @staticmethod
    def _get_keys_dir(cert_file_path: str) -> str:
        """Get keys directory from certificate file path.

        :param cert_file_path: Path to certificate file
        :return: Keys directory path
        """
        directory, _ = os.path.split(cert_file_path)
        if directory.endswith("crts"):
            return directory[:-4] + "keys"
        return cert_file_path


class CsfBuildDirector:
    """CSF command build director."""

    def __init__(self, builder: CsfBuilder) -> None:
        """CSF build director class constructor.

        :param builder: CSF builder
        """
        self._builder: CsfBuilder = builder

    @property
    def builder(self) -> CsfBuilder:
        """CSF builder property."""
        return self._builder

    def build_csf(self) -> None:
        """Build individual CSF commands."""
        self._builder.build_csf_header()
        self._builder.build_csf_install_srk()
        self._builder.build_csf_install_csfk()
        self._builder.build_authenticate_csfk()
        self._builder.build_install_key_csfk()
        self._builder.build_authenticate_data()
        self._builder.build_set_engine()
        self._builder.build_unlock_engine()
        self._builder.build_install_secret_key()
        self._builder.build_decrypt_data()
        self._builder.finish()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Translator is responsible for converting stringified keys into values."""

import logging
from typing import Optional

from .exceptions import SPSDKPfrcMissingConfigError
from .pfr import CFPA, CMPA

logger = logging.getLogger(__name__)


class Translator:
    """Translates single strings (register/key names) into values."""

    def __init__(
        self,
        cmpa: Optional[CMPA] = None,
        cfpa: Optional[CFPA] = None,
    ) -> None:
        """Initialize CMPA and CFPA data.

        :param cmpa: configuration data loaded from CMPA config file
        :param cfpa: configuration data loaded from CFPA config file
        """
        self.cmpa_obj = cmpa
        self.cfpa_obj = cfpa
        self.handlers = {
            "CMPA": self._cmpa_translate,
            "CFPA": self._cfpa_translate,
            "UTIL": self._util_translate,
        }

    def translate(self, key: str) -> int:
        """Lookup register's (or generic key's) value.

        :param key: Register's (key's) stringified name
        :return: Register's (key's) value
        :raises SPSDKPfrcTranslationError: Raises when the configuration for given key is not defined
        """
        area, value = key.split(".", maxsplit=1)
        logger.debug(f"Area designator: {area}")
        logger.debug(f"Register designator: {value}")
        return self.handlers[area](value)

    def _cmpa_translate(self, key: str) -> int:
        """Handler for CMPA data."""
        if not self.cmpa_obj:
            raise SPSDKPfrcMissingConfigError(f"Cannot translate {key}. CMPA config not defined")
        logger.debug(f"Extracting value from {key}")
        splitted = key.split(".", maxsplit=1)
        register = self.cmpa_obj.registers.find_reg(splitted[0])
        if len(splitted) == 2:
            value = register.find_bitfield(splitted[1]).get_value()
        else:
            value = register.get_value()
        logger.debug(f"Extracted value {value:x}")

        return value

    def _cfpa_translate(self, key: str) -> int:
        """Handler for CFPA data."""
        if not self.cfpa_obj:
            raise SPSDKPfrcMissingConfigError(f"Cannot translate {key}. CFPA config not defined")
        logger.debug(f"Extracting value from {key}")
        splitted = key.split(".", maxsplit=1)
        register = self.cfpa_obj.registers.find_reg(splitted[0])
        if len(splitted) == 2:
            value = register.find_bitfield(splitted[1]).get_value()
        else:
            value = register.get_value()
        logger.debug(f"Extracted value {value:x}")

        return value

    def _util_translate(self, key: str) -> int:  # pylint: disable=no-self-use
        """Handler for Utils data."""
        values = {"isUDSKeyCodeValid": False}
        return values[key]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""NXP MCU Image tool."""
import datetime
import logging
import os
import sys
from binascii import unhexlify
from typing import List, Optional

import click

from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroup,
    spsdk_apps_common_options,
    spsdk_plugin_option,
)
from spsdk.apps.utils.utils import (
    INT,
    SPSDKAppError,
    catch_spsdk_error,
    filepath_from_config,
    get_key,
    store_key,
)
from spsdk.crypto.signature_provider import get_signature_provider
from spsdk.exceptions import SPSDKError
from spsdk.image import TrustZone, get_mbi_class
from spsdk.image.ahab.ahab_container import AHABImage
from spsdk.image.ahab.signed_msg import SignedMessage
from spsdk.image.bee import BeeNxp
from spsdk.image.bootable_image.bimg import BootableImage, get_bimg_class
from spsdk.image.fcb.fcb import FCB
from spsdk.image.hab.config_parser import ImageConfig
from spsdk.image.hab.hab_container import HabContainer
from spsdk.image.keystore import KeyStore
from spsdk.image.mbimg import (
    MasterBootImage,
    mbi_generate_config_templates,
    mbi_get_supported_families,
)
from spsdk.image.xmcd.xmcd import XMCD
from spsdk.sbfile.sb2 import sly_bd_parser as bd_parser
from spsdk.sbfile.sb2.commands import CmdLoad
from spsdk.sbfile.sb2.images import BootImageV21, generate_SB21
from spsdk.sbfile.sb31.images import SB3_SCH_FILE, SecureBinary31
from spsdk.utils.crypto.cert_blocks import CertBlockV2, CertBlockV31
from spsdk.utils.crypto.iee import IeeNxp
from spsdk.utils.crypto.otfad import OtfadNxp
from spsdk.utils.images import BinaryImage, BinaryPattern
from spsdk.utils.misc import (
    align,
    align_block,
    get_abs_path,
    load_binary,
    load_configuration,
    load_text,
    value_to_int,
    write_file,
)
from spsdk.utils.plugins import load_plugin_from_source
from spsdk.utils.schema_validator import (
    CommentedConfig,
    ConfigTemplate,
    ValidationSchemas,
    check_config,
)

logger = logging.getLogger(__name__)


@click.group(name="nxpimage", no_args_is_help=True, cls=CommandsTreeGroup)
@spsdk_apps_common_options
def main(log_level: int) -> None:
    """NXP Image tool.

    Manage various kinds of images for NXP parts.
    It's successor of obsolete ELFTOSB tool.
    """
    spsdk_logger.install(level=log_level)


@main.group(name="mbi", no_args_is_help=True)
def mbi_group() -> None:
    """Group of sub-commands related to Master Boot Images."""


@mbi_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
@spsdk_plugin_option
def mbi_export_command(
    config: str,
    plugin: str,
) -> None:
    """Generate Master Boot Image from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-templates'.
    """
    mbi_export(config, plugin)


def mbi_export(config: str, plugin: Optional[str] = None) -> None:
    """Generate Master Boot Image from YAML/JSON configuration.

    :param config: Path to the YAML/JSON configuration
    :param plugin: Path to external python file containing a custom SignatureProvider implementation.
    """
    config_data = load_configuration(config)
    if plugin:
        load_plugin_from_source(plugin)
    config_dir = os.path.dirname(config)
    mbi_cls = get_mbi_class(config_data)
    check_config(config_data, mbi_cls.get_validation_schemas(), search_paths=[config_dir])
    mbi_obj = mbi_cls()
    mbi_obj.load_from_config(config_data, search_paths=[config_dir])
    mbi_data = mbi_obj.export()

    mbi_output_file_path = get_abs_path(config_data["masterBootOutputFile"], config_dir)
    write_file(mbi_data, mbi_output_file_path, mode="wb")

    click.echo(f"Success. (Master Boot Image: {mbi_output_file_path} created.)")


@mbi_group.command(name="parse", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="lpc55s3x",
    type=click.Choice(mbi_get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary MBI image to parse.",
)
@click.option(
    "-k",
    "--dek",
    type=str,
    required=False,
    help=(
        "Data encryption key, if it's specified, the parse method tries decrypt all encrypted images. "
        "It could be specified as binary/HEX text file path or directly HEX string"
    ),
)
@click.argument("parsed-data", type=click.Path(resolve_path=True))
def mbi_parse_command(chip_family: str, binary: str, dek: str, parsed_data: str) -> None:
    """Parse MBI Image into YAML configuration and binary images."""
    mbi_parse(chip_family, binary, dek, parsed_data)


def mbi_parse(chip_family: str, binary: str, dek: str, parsed_data: str) -> None:
    """Parse AHAB Image into YAML configuration and binary images."""
    if not os.path.exists(parsed_data):
        os.makedirs(parsed_data, exist_ok=True)

    mbi = MasterBootImage.parse(family=chip_family, data=load_binary(binary), dek=dek)

    if not mbi:
        click.echo(f"Failed. (MBI: {binary} parsing failed.)")
        return

    mbi.create_config(output_folder=parsed_data)

    click.echo(f"Success. (MBI: {binary} has been parsed and stored into {parsed_data} )")


@mbi_group.command(name="get-templates", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="lpc55s3x",
    type=click.Choice(mbi_get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path())
def mbi_get_templates_command(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of MBI configurations in YAML format.

    The path for template files is specified as argument of this command.
    """
    mbi_get_templates(chip_family, overwrite, output)


def mbi_get_templates(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of MBI configurations in YAML format."""
    templates = mbi_generate_config_templates(chip_family)
    for file_name, template in templates.items():
        full_file_name = os.path.abspath(os.path.join(output, file_name + ".yaml"))
        if not os.path.isfile(full_file_name) or overwrite:
            click.echo(f"Creating {full_file_name} template file.")
            write_file(template, full_file_name)
        else:
            click.echo(f"Skip creating {full_file_name}, this file already exists.")


@main.group(name="sb21", no_args_is_help=True)
def sb21_group() -> None:
    """Group of sub-commands related to Secure Binary 2.1."""


@sb21_group.command(name="export", no_args_is_help=True)
@click.option(
    "-c",
    "--command",
    type=click.Path(exists=True),
    help="BD configuration file to produce secure binary v2.x",
)
@click.option("-o", "--output", type=click.Path(resolve_path=True), help="Output file path.")
@click.option(
    "-k", "--key", type=click.Path(exists=True), help="Add a key file and enable encryption."
)
@click.option(
    "-s",
    "--pkey",
    type=str,
    help="Path to private key or signature provider configuration used for signing.",
)
@click.option(
    "-S",
    "--cert",
    type=click.Path(exists=True),
    multiple=True,
    help="Path to certificate files for signing. The first certificate will be \
the self signed root key certificate.",
)
@click.option(
    "-R",
    "--root-key-cert",
    type=click.Path(exists=True),
    multiple=True,
    help="Path to root key certificate file(s) for verifying other certificates. \
Only 4 root key certificates are allowed, others are ignored. \
One of the certificates must match the first certificate passed \
with -S/--cert arg.",
)
@click.option(
    "-h",
    "--hash-of-hashes",
    type=click.Path(),
    help="Path to output hash of hashes of root keys. If argument is not \
provided, then by default the tool creates hash.bin in the working directory.",
)
@click.option(
    "--plugin",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="External python file containing a custom SignatureProvider implementation.",
)
@click.argument("external", type=click.Path(), nargs=-1)
def sb21_export_command(
    command: str,
    output: str,
    key: str,
    pkey: str,
    cert: List[str],
    root_key_cert: List[str],
    hash_of_hashes: str,
    plugin: str,
    external: List[str],
) -> None:
    """Generate Secure Binary v2.1 Image from configuration.

    EXTERNAL is a space separated list of external binary files defined in BD file
    """
    sb21_export(command, output, key, pkey, cert, root_key_cert, hash_of_hashes, plugin, external)


def sb21_export(
    command: str,
    output: str,
    key: str,
    pkey: str,
    cert: List[str],
    root_key_cert: List[str],
    hash_of_hashes: str,
    plugin: str,
    external: List[str],
) -> None:
    """Generate Secure Binary v2.1 Image from configuration."""
    if plugin:
        load_plugin_from_source(plugin)
    if output is None:
        raise SPSDKAppError("Error: no output file was specified")
    signature_provider = (
        get_signature_provider(local_file_key=pkey)
        if os.path.isfile(pkey)
        else get_signature_provider(sp_cfg=pkey)
    )

    try:
        sb2_data = generate_SB21(
            bd_file_path=command,
            key_file_path=key,
            signature_provider=signature_provider,
            signing_certificate_file_paths=cert,
            root_key_certificate_paths=root_key_cert,
            hoh_out_path=hash_of_hashes,
            external_files=external,
        )
        write_file(sb2_data, output, mode="wb")
    except SPSDKError as exc:
        raise SPSDKAppError(f"The SB2.1 file generation failed: ({str(exc)}).") from exc
    else:
        click.echo(f"Success. (Secure binary 2.1: {output} created.)")


@sb21_group.command(name="parse")
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to the SB2 container that would be parsed.",
)
@click.option(
    "-k",
    "--key",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Key file for SB2 decryption in plaintext",
)
@click.argument("parsed-data", type=click.Path())
def sb21_parse_command(binary: str, key: str, parsed_data: str) -> None:
    """Parse Secure Binary v2.1 Image."""
    sb21_parse(binary, key, parsed_data)


def sb21_parse(binary: str, key: str, parsed_data: str) -> None:
    """Parse Secure Binary v2.1 Image."""
    # transform text-based KEK into bytes
    sb_kek = unhexlify(load_text(key))

    try:
        parsed_sb = BootImageV21.parse(data=load_binary(binary), kek=sb_kek)
    except SPSDKError as exc:
        logger.debug(f"SB21 parse: Attempt to parse image failed: {str(exc)}")

    if isinstance(parsed_sb.cert_block, CertBlockV2):
        for cert_idx, certificate in enumerate(parsed_sb.cert_block.certificates):
            file_name = os.path.abspath(
                os.path.join(parsed_data, f"certificate_{cert_idx}_der.cer")
            )
            logger.debug(f"Dumping certificate {file_name}")
            write_file(certificate.dump(), file_name, mode="wb")

    for section_idx, boot_sections in enumerate(parsed_sb.boot_sections):
        for command_idx, command in enumerate(boot_sections._commands):
            if isinstance(command, CmdLoad):
                file_name = os.path.abspath(
                    os.path.join(
                        parsed_data, f"section_{section_idx}_load_command_{command_idx}_data.bin"
                    )
                )
                logger.debug(f"Dumping load command data {file_name}")
                write_file(command.data, file_name, mode="wb")

    logger.debug(parsed_sb.info())
    write_file(
        parsed_sb.info(),
        os.path.join(parsed_data, "parsed_info.txt"),
    )
    click.echo(f"Success. (SB21: {binary} has been parsed and stored into {parsed_data}.)")
    click.echo(
        "Please note that the exported binary images from load command might contain padding"
    )


@sb21_group.command(name="get-sbkek", no_args_is_help=False)
@click.option(
    "-k",
    "--master-key",
    type=str,
    help="AES-256 master key as hexadecimal string or path to file containing key in plain text or in binary",
)
@click.option(
    "-o",
    "--output-folder",
    type=click.Path(resolve_path=True),
    help="Output folder where the sbkek.txt and sbkek.bin will be stored",
)
def get_sbkek_command(master_key: str, output_folder: str) -> None:
    """Compute SBKEK (AES-256) value and optionally store it as plain text and as binary.

    SBKEK is AES-256 symmetric key used for encryption and decryption of SB.
    Plain text version is used for SB generation.
    Binary format is to be written to the keystore.
    The same format is also used for USERKEK.

    For OTP, the SBKEK is derived from OTP master key:
    SB2_KEK = AES256(OTP_MASTER_KEY,
    03000000_00000000_00000000_00000000_04000000_00000000_00000000_00000000)

    Master key is not needed when using PUF as key storage

    The computed SBKEK is shown as hexadecimal text on STDOUT,
    SBKEK is stored in plain text and in binary if the 'output-folder' is specified,
    """
    get_sbkek(master_key, output_folder)


def get_sbkek(master_key: str, output_folder: str) -> None:
    """Compute SBKEK (AES-256) value and optionally store it as plain text and as binary."""
    otp_master_key = get_key(master_key, KeyStore.OTP_MASTER_KEY_SIZE)
    sbkek = KeyStore.derive_sb_kek_key(otp_master_key)

    click.echo(f"SBKEK: {sbkek.hex()}")
    click.echo(f"(OTP) MASTER KEY: {otp_master_key.hex()}")

    if output_folder:
        store_key(os.path.join(output_folder, "sbkek"), sbkek, reverse=True)
        store_key(os.path.join(output_folder, "otp_master_key"), otp_master_key)
        click.echo(f"Keys have been stored to: {output_folder}")


@main.group(name="sb31")
def sb31_group() -> None:
    """Group of sub-commands related to Secure Binary 3.1."""


@sb31_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def sb31_export_command(config: str) -> None:
    """Generate Secure Binary v3.1 Image from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration.
    RoTKTH/SB3KDK is printed out in verbose mode.

    The configuration template files could be generated by subcommand 'get-template'.
    """
    sb31_export(config)


def sb31_export(config: str) -> None:
    """Generate Secure Binary v3.1 Image from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    check_config(config_data, SecureBinary31.get_validation_schemas_family())
    schemas = SecureBinary31.get_validation_schemas(
        config_data["family"], include_test_configuration=True
    )
    schemas.append(ValidationSchemas.get_schema_file(SB3_SCH_FILE)["sb3_output"])
    check_config(config_data, schemas, search_paths=[config_dir])
    sb3 = SecureBinary31.load_from_config(config_data, search_paths=[config_dir])

    sb3_data = sb3.export()
    sb3_output_file_path = get_abs_path(config_data["containerOutputFile"], config_dir)
    write_file(sb3_data, sb3_output_file_path, mode="wb")

    click.echo(f"Success. (Secure binary 3.1: {sb3_output_file_path} created.)")


@sb31_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="lpc55s3x",
    type=click.Choice(SecureBinary31.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def sb31_get_template_command(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    sb31_get_template(chip_family, overwrite, output)


def sb31_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(
            SecureBinary31.generate_config_template(chip_family)[f"{chip_family}_sb31"], output
        )
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="cert-block", no_args_is_help=True)
def cert_block_group() -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to certification block."""


@cert_block_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def cert_block_get_template_command(overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    cert_block_get_template(overwrite, output)


def cert_block_get_template(overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(CertBlockV31.generate_config_template(), output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@cert_block_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def cert_block_export_command(config: str) -> None:
    """Generate Certificate Block v3.1 from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration.
    RoTKTH is printed out in verbose mode.

    The configuration template files could be generated by subcommand 'get-template'.
    """
    cert_block_export(config)


def cert_block_export(config: str) -> None:
    """Generate Certificate Block v3.1 from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    schemas = CertBlockV31.get_validation_schemas()
    check_config(config_data, schemas, search_paths=[config_dir])
    cert_block = CertBlockV31.from_config(config_data, search_paths=[config_dir])
    cert_data = cert_block.export()

    cert_block_output_file_path = get_abs_path(config_data["containerOutputFile"], config_dir)
    write_file(cert_data, cert_block_output_file_path, mode="wb")

    click.echo(f"Success. (Certificate Block 3.1: {cert_block_output_file_path} created.)")


@cert_block_group.command(name="parse", no_args_is_help=True)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary Certificate v3.1 image to parse.",
)
@click.argument("parsed-data", type=click.Path(resolve_path=True))
def cert_block_parse_command(binary: str, parsed_data: str) -> None:
    """Parse Certificate Block v3.1.

    PARSED-DATA is path where the parsed certification block will be stored.
    RoTKTH is printed out in verbose mode.
    """
    cert_block_parse(binary, parsed_data)


def cert_block_parse(binary: str, parsed_data: str) -> None:
    """Parse Certificate Block v3.1."""
    cert_block = CertBlockV31.parse(load_binary(binary))
    logger.info(cert_block.info())
    write_file(
        cert_block.create_config(parsed_data), os.path.join(parsed_data, "cert_block_config.yaml")
    )
    click.echo(f"Success. (Certificate Block 3.1: {binary} has been parsed into {parsed_data}.)")


@main.group(name="tz", no_args_is_help=True)
def tz_group() -> None:
    """Group of sub-commands related to Trust Zone."""


@tz_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def tz_export_command(config: str) -> None:
    """Generate TrustZone Image from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    tz_export(config)


def tz_export(config: str) -> None:
    """Generate TrustZone Image from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    check_config(config_data, TrustZone.get_validation_schemas_family())
    check_config(
        config_data,
        TrustZone.get_validation_schemas(config_data["family"]),
        search_paths=[config_dir],
    )
    trust_zone = TrustZone.from_config(config_data)
    tz_data = trust_zone.export()
    output_file = get_abs_path(config_data["tzpOutputFile"], config_dir)
    write_file(tz_data, output_file, mode="wb")
    click.echo(f"Success. (Trust Zone binary: {output_file} created.)")


@tz_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="lpc55s3x",
    type=click.Choice(TrustZone.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-r",
    "--revision",
    type=str,
    default="latest",
    help="Chip revision; if not specified, most recent one will be used",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def tz_get_template_command(chip_family: str, revision: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    tz_get_template(chip_family, revision, overwrite, output)


def tz_get_template(chip_family: str, revision: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(
            TrustZone.generate_config_template(chip_family, revision)[f"{chip_family}_tz"], output
        )
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="ahab", no_args_is_help=True)
def ahab_group() -> None:
    """Group of sub-commands related to AHAB."""


@ahab_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def ahab_export_command(config: str) -> None:
    """Generate AHAB Image from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    ahab_export(config)


def ahab_export(config: str) -> None:
    """Generate AHAB Image from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    schemas = AHABImage.get_validation_schemas()
    check_config(config_data, schemas, search_paths=[config_dir])
    ahab_cnt = AHABImage.load_from_config(config_data, search_paths=[config_dir])
    ahab_data = ahab_cnt.export()

    ahab_output_file_path = get_abs_path(config_data["output"], config_dir)
    write_file(ahab_data, ahab_output_file_path, mode="wb")

    logger.info(f"Created AHAB Image:\n{ahab_cnt.image_info().info()}")
    logger.info(f"Created AHAB Image memory map:\n{ahab_cnt.image_info().draw()}")
    click.echo(f"Success. (AHAB: {ahab_output_file_path} created.)")

    ahab_output_dir, ahab_output_file = os.path.split(ahab_output_file_path)
    ahab_output_file_no_ext, _ = os.path.splitext(ahab_output_file)
    for cnt_ix, container in enumerate(ahab_cnt.ahab_containers):
        if container.flag_srk_set == "nxp":
            logger.debug("Skipping generating hashes for NXP container")
            continue
        srk_table = container.signature_block.srk_table
        file_name = f"{ahab_output_file_no_ext}_{container.flag_srk_set}{cnt_ix}_srk_hash"
        if srk_table:
            srkh = srk_table.compute_srk_hash()
            write_file(srkh.hex().upper(), get_abs_path(f"{file_name}.txt", ahab_output_dir))
            try:
                blhost_script = ahab_cnt.create_srk_hash_blhost_script(cnt_ix)
                write_file(blhost_script, get_abs_path(f"{file_name}_blhost.bcf", ahab_output_dir))
            except SPSDKError:
                pass
            click.echo(f"Generated SRK hash files ({os.path.abspath(file_name)}*.*).")


@ahab_group.command(name="parse", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt118x",
    type=click.Choice(AHABImage.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-t",
    "--image-type",
    type=click.Choice(AHABImage.IMAGE_TYPES),
    help="AHAB image type., default is XIP.",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary AHAB image to parse.",
)
@click.option(
    "-k",
    "--dek",
    type=str,
    required=False,
    help=(
        "Data encryption key, if it's specified, the parse method tries decrypt all encrypted images. "
        "It could be specified as binary/HEX text file path or directly HEX string"
    ),
)
@click.argument("parsed-data", type=click.Path(resolve_path=True))
def ahab_parse_command(
    chip_family: str, image_type: str, binary: str, dek: str, parsed_data: str
) -> None:
    """Parse AHAB Image into YAML configuration and binary images."""
    ahab_parse(chip_family, image_type, binary, dek, parsed_data)


def ahab_parse(chip_family: str, image_type: str, binary: str, dek: str, parsed_data: str) -> None:
    """Parse AHAB Image into YAML configuration and binary images."""
    if not os.path.exists(parsed_data):
        os.makedirs(parsed_data, exist_ok=True)
    image_types = [image_type] if image_type else AHABImage.IMAGE_TYPES
    for img_type in image_types:
        try:
            ahab_image = AHABImage(family=chip_family, image_type=img_type)
            ahab_image.parse(load_binary(binary))
            ahab_image.update_fields()
            ahab_image.validate()
        except SPSDKError as exc:
            logger.debug(f"AHAB parse: Attempt to parse image as {img_type} failed: {str(exc)}")
            ahab_image = None
        else:
            assert ahab_image
            if (
                ahab_image.get_first_data_image_address() < ahab_image.data_images_start
                and img_type != "serial_downloader"
            ):
                ahab_image = None
            else:
                break

    if not ahab_image:
        click.echo(f"Failed. (AHAB: {binary} parsing failed.)")
        return

    logger.info(f"Identified AHAB image type: {img_type}")
    logger.info(f"Parsed AHAB image memory map: {ahab_image.image_info().draw()}")
    if dek:
        for container in ahab_image.ahab_containers:
            if container.flag_srk_set != "nxp" and container.signature_block.blob:
                container.signature_block.blob.dek = get_key(
                    dek, container.signature_block.blob._size // 8
                )
                container.decrypt_data()

    config = ahab_image.create_config(parsed_data)
    write_file(
        CommentedConfig(
            main_title=(
                f"AHAB recreated configuration from :"
                f"{datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}."
            ),
            schemas=AHABImage.get_validation_schemas(),
            values=config,
            export_template=False,
        ).export_to_yaml(),
        os.path.join(parsed_data, "parsed_config.yaml"),
    )
    click.echo(f"Success. (AHAB: {binary} has been parsed and stored into {parsed_data}.)")

    for cnt_ix, container in enumerate(ahab_image.ahab_containers):
        srk_table = container.signature_block.srk_table
        file_name = os.path.join(parsed_data, f"{container.flag_srk_set}{cnt_ix}_srk_hash")
        if srk_table:
            srkh = srk_table.compute_srk_hash()
            write_file(srkh.hex().upper(), f"{file_name}.txt")
            try:
                blhost_script = ahab_image.create_srk_hash_blhost_script(cnt_ix)
                write_file(blhost_script, f"{file_name}_blhost.bcf")
            except SPSDKError:
                pass
            click.echo(f"Generated SRK hash files ({file_name}*.*).")


@ahab_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt118x",
    type=click.Choice(AHABImage.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def ahab_get_template_command(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    ahab_get_template(chip_family, overwrite, output)


def ahab_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(AHABImage.generate_config_template(chip_family)[f"{chip_family}_ahab"], output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(no_args_is_help=True)
def signed_msg() -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to Signed messages."""


@signed_msg.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def signed_msg_export(config: str) -> None:
    """Generate Signed message Image from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    schemas = SignedMessage.get_validation_schemas()
    check_config(config_data, schemas, search_paths=[config_dir])
    smsg = SignedMessage.load_from_config(config_data, search_paths=[config_dir])

    signed_msg_data = smsg.export()

    signed_msg_output_file_path = get_abs_path(config_data["output"], config_dir)
    write_file(signed_msg_data, signed_msg_output_file_path, mode="wb")

    logger.info(f"Created Signed message Image:\n{smsg.image_info().info()}")
    logger.info(f"Created Signed message Image memory map:\n{smsg.image_info().draw()}")
    click.echo(f"Success. (Signed message: {signed_msg_output_file_path} created.)")


@signed_msg.command(name="parse", no_args_is_help=True)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary Signed message image to parse.",
)
@click.argument("parsed-data", type=click.Path(resolve_path=True))
def signed_msg_parse(binary: str, parsed_data: str) -> None:
    """Parse Signed message Image into YAML configuration and binary images."""
    if not os.path.exists(parsed_data):
        os.makedirs(parsed_data, exist_ok=True)
    try:
        signed_message = SignedMessage.parse(load_binary(binary))
        signed_message.update_fields()
        signed_message.validate({})
    except SPSDKError as exc:
        click.echo(f"Signed message parsing failed: {binary} ,({str(exc)})")
        return

    logger.info(f"Parsed Signed message image memory map: {signed_message.image_info().draw()}")
    config = signed_message.create_config(parsed_data)
    write_file(
        ConfigTemplate.convert_cm_to_yaml(config),
        os.path.join(parsed_data, "parsed_config.yaml"),
    )
    click.echo(
        f"Success. (Signed message: {binary} has been parsed and stored into {parsed_data}.)"
    )

    srk_table = signed_message.signature_block.srk_table
    file_name = os.path.join(parsed_data, f"{signed_message.flag_srk_set}_srk_hash")
    if srk_table:
        srkh = srk_table.compute_srk_hash()
        write_file(srkh.hex().upper(), f"{file_name}.txt")
        click.echo(f"Generated SRK hash files ({file_name}*.*).")


@signed_msg.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt118x",
    type=click.Choice(AHABImage.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def signed_msg_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(
            SignedMessage.generate_config_template(chip_family)[f"{chip_family}_signed_msg"], output
        )
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="otfad", no_args_is_help=True)
def otfad_group() -> None:
    """Group of sub-commands related to OTFAD."""


@otfad_group.command(name="export", no_args_is_help=True)
@click.option(
    "-a",
    "--alignment",
    default="512",
    type=INT(),
    help="Alignment of key blob data blocks. Default value is 512 B if not specified.",
    required=False,
)
@click.option(
    "-i",
    "--index",
    type=INT(),
    help=(
        "OTFAD peripheral index - This is needed to generate proper "
        "indexes of fuses in optional BLHOST script. If not specified, BLHOST fuse script won't be generated"
    ),
    required=False,
)
@click.argument("config", type=click.Path(exists=True, readable=True))
def otfad_export_command(alignment: int, config: str, index: Optional[int] = None) -> None:
    """Generate OTFAD Images from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    otfad_export(alignment, config, index)


def otfad_export(alignment: int, config: str, index: Optional[int] = None) -> None:
    """Generate OTFAD Images from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    check_config(config_data, OtfadNxp.get_validation_schemas_family(), search_paths=[config_dir])
    family = config_data["family"]
    schemas = OtfadNxp.get_validation_schemas(family)
    check_config(config_data, schemas, search_paths=[config_dir])
    otfad = OtfadNxp.load_from_config(config_data, config_dir, search_paths=[config_dir])

    output_folder = get_abs_path(config_data["output_folder"], config_dir)
    otfad_table_name = filepath_from_config(
        config_data, "keyblob_name", "OTFAD_Table", config_dir, output_folder
    )
    binary_image = otfad.binary_image(data_alignment=alignment, otfad_table_name=otfad_table_name)
    logger.info(f" The OTFAD image structure:\n{binary_image.draw()}")
    otfad_all = filepath_from_config(
        config_data, "output_name", "otfad_whole_image", config_dir, output_folder
    )
    if otfad_all != "":
        write_file(binary_image.export(), otfad_all, mode="wb")
        logger.info(f"Created OTFAD Image:\n{otfad_all}")
    else:
        logger.info("Skipping export of OTFAD image")
    sb21_supported = otfad.database.get_device_value("sb_21_supported", otfad.family, default=False)
    memory_map = (
        "In folder is stored two kind of files:\n"
        "  -  Binary file that contains whole image data including "
        "OTFAD table and key blobs data 'otfad_whole_image.bin'.\n"
    )
    if sb21_supported:
        memory_map += "  -  Example of BD file to simplify creating the SB2.1 file from the OTFAD source files.\n"
        bd_file_sources = "sources {"
        bd_file_section0 = "section (0) {"

    memory_map += (
        "  -  Set of separated binary files, one with OTFAD table, and one for each used key blob.\n"
        f"\nOTFAD memory map:\n{binary_image.draw(no_color=True)}"
    )

    for i, image in enumerate(binary_image.sub_images):
        if image.name != "":
            write_file(image.export(), image.name, mode="wb")
            logger.info(f"Created OTFAD Image:\n{image.name}")
            memory_map += f"\n{image.name}:\n{image.info()}"
        else:
            logger.info(
                f"Skipping export of {image.info()}, value is blank in the configuration file"
            )
        if sb21_supported:
            bd_file_sources += f'\n    image{i} = "{image.name}";'
            bd_file_section0 += f"\n    // Load Image: {image.name}"
            bd_file_section0 += f"\n    erase {hex(image.absolute_address)}..{hex(image.absolute_address+len(image))};"
            bd_file_section0 += f"\n    load image{i} > {hex(image.absolute_address)}"

    readme_file = os.path.join(output_folder, "readme.txt")

    if config_data.get("generate_readme", True):
        write_file(memory_map, readme_file)
        logger.info(f"Created OTFAD readme file:\n{readme_file}")
    else:
        logger.info("Skipping generation of OTFAD readme file")

    if sb21_supported:
        bd_file_name = os.path.join(output_folder, "sb21_otfad_example.bd")
        bd_file_sources += "\n}\n"
        bd_file_section0 += "\n}\n"
        bd_file = (
            "options {\n"
            "    flags = 0x8; // for sb2.1 use only 0x8 encrypted + signed\n"
            "    buildNumber = 0x1;\n"
            '    productVersion = "1.00.00";\n'
            '    componentVersion = "1.00.00";\n'
            '    secureBinaryVersion = "2.1";\n'
            "}\n"
        )
        bd_file += bd_file_sources
        bd_file += bd_file_section0

        write_file(bd_file, bd_file_name)
        logger.info(f"Created OTFAD BD file example:\n{bd_file_name}")

    if otfad.database.get_device_value("has_kek_fuses", otfad.family, default=False) and index:
        blhost_script = None
        blhost_script = otfad.get_blhost_script_otp_kek(index)
        if blhost_script:
            blhost_script_filename = os.path.join(
                output_folder, f"otfad{index}_{otfad.family}_blhost.bcf"
            )
            write_file(blhost_script, blhost_script_filename)
            click.echo(f"Created OTFAD BLHOST load fuses script:\n{blhost_script_filename}")

    click.echo("Success. OTFAD files have been created")


@otfad_group.command(name="get-kek", no_args_is_help=False)
@click.option(
    "-m",
    "--otp-master-key",
    type=str,
    help="OTP MASTER KEY in hexadecimal format or file name to binary or text file with key.",
)
@click.option(
    "-k",
    "--otfad-key",
    type=str,
    help="OTFAD KEY (OTFAD KEK seed) in hexadecimal format or file name to binary or text file with key.",
)
@click.option(
    "-f",
    "--chip-family",
    type=click.Choice(OtfadNxp.get_supported_families()),
    help=(
        "Optional family, if specified, the tool generates the BLHOST scripts to load key fuses."
        " To use this feature, the '-o' options has to be also defined!"
    ),
)
@click.option(
    "-o",
    "--output-folder",
    type=click.Path(resolve_path=True),
    help="Optional result output folder (otfad_kek.bin/txt, optionally BLHOST scripts to load keys into Fuses)",
)
def otfad_get_kek_command(
    otp_master_key: str, otfad_key: str, chip_family: str, output_folder: str
) -> None:
    """Compute OTFAD KEK value and optionally store it into folder in various formats.

    The computed OTFAD KEK is shown in hexadecimal text, if the 'output-folder' is specified,
    it is stored in folder also in binary format.
    """
    otfad_get_kek(otp_master_key, otfad_key, chip_family, output_folder)


def otfad_get_kek(
    otp_master_key: str, otfad_key: str, chip_family: str, output_folder: str
) -> None:
    """Compute OTFAD KEK value and optionally store it into folder in various formats."""
    omk = get_key(otp_master_key, KeyStore.OTP_MASTER_KEY_SIZE)
    ok = get_key(otfad_key, KeyStore.OTFAD_KEY_SIZE)  # pylint:disable=invalid-name

    otfad_kek = KeyStore.derive_otfad_kek_key(omk, ok)

    click.echo(f"OTP MASTER KEY: {omk.hex()}")
    click.echo(f"OTFAD KEY:      {ok.hex()}")
    click.echo(f"OTFAD KEK:      {otfad_kek.hex()}")

    blhost_script = None
    if chip_family and chip_family in OtfadNxp.get_supported_families():
        blhost_script = OtfadNxp.get_blhost_script_otp_keys(
            chip_family, otp_master_key=omk, otfad_key_seed=ok
        )
        if not output_folder:
            click.echo(f"OTFAD BLHOST load fuses script:\n{blhost_script}")

    if output_folder:
        store_key(os.path.join(output_folder, "otp_master_key"), omk)
        store_key(os.path.join(output_folder, "otfad_key"), ok)
        store_key(os.path.join(output_folder, "otfad_kek"), otfad_kek)
        if blhost_script:
            write_file(blhost_script, os.path.join(output_folder, "otfad_otp_fuses.bcf"))
        click.echo(f"Result files has been stored into: {output_folder}")


@otfad_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    type=click.Choice(OtfadNxp.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def otfad_get_template_command(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    otfad_get_template(chip_family, overwrite, output)


def otfad_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(OtfadNxp.generate_config_template(chip_family)[f"{chip_family}_otfad"], output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="iee", no_args_is_help=True)
def iee_group() -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to IEE."""


@iee_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def iee_export_command(config: str) -> None:
    """Generate IEE Images from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    iee_export(config)


def iee_export(config: str) -> None:
    """Generate IEE Images from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    check_config(config_data, IeeNxp.get_validation_schemas_family(), search_paths=[config_dir])
    family = config_data["family"]
    schemas = IeeNxp.get_validation_schemas(family)
    check_config(config_data, schemas, search_paths=[config_dir])
    iee = IeeNxp.load_from_config(config_data, config_dir, search_paths=[config_dir])

    output_folder = get_abs_path(config_data["output_folder"], config_dir)
    iee_all = filepath_from_config(config_data, "output_name", "iee_whole_image", output_folder)
    keyblob_name = filepath_from_config(
        config_data, "keyblob_name", "iee_keyblob", config_dir, output_folder
    )

    binary_image = iee.binary_image(keyblob_name=keyblob_name, image_name=iee_all)
    logger.info(binary_image.draw())

    if iee_all == "":
        logger.info("Skipping export of IEE whole image")
    else:
        write_file(binary_image.export(), iee_all, mode="wb")
        click.echo(f"Created IEE Image containing keyblobs and encrypted data:\n{iee_all}")

    memory_map = (
        "Output folder contains:\n"
        "  -  Binary file that contains whole image data including "
        f"IEE key blobs data {iee_all}.\n"
        f"IEE memory map:\n{binary_image.draw(no_color=True)}"
    )

    for image in binary_image.sub_images:
        if image.name != "":
            write_file(image.export(), image.name, mode="wb")
            logger.info(f"Created Encrypted IEE data blob {image.description}:\n{image.name}")
            memory_map += f"\n{image.name}:\n{image.info()}"
        else:
            logger.info(
                f"Skipping export of {image.info()}, value is blank in the configuration file"
            )

    readme_file = os.path.join(output_folder, "readme.txt")

    if config_data.get("generate_readme", True):
        write_file(memory_map, readme_file)
        logger.info(f"Created IEE readme file:\n{readme_file}")
    else:
        logger.info("Skipping generation of IEE readme file")

    if iee.database.get_device_value("has_kek_fuses", iee.family) and config_data.get(
        "generate_fuses_script"
    ):
        blhost_script = iee.get_blhost_script_otp_kek()
        blhost_script_filename = os.path.join(output_folder, f"iee_{iee.family}_blhost.bcf")
        write_file(blhost_script, blhost_script_filename)
        click.echo(f"Created IEE BLHOST load fuses script:\n{blhost_script_filename}")
    else:
        logger.info("Skipping generation of IEE BLHOST load fuses script")

    click.echo("Success. IEE files have been created")


@iee_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt117x",
    type=click.Choice(IeeNxp.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def iee_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(IeeNxp.generate_config_template(chip_family)[f"{chip_family}_iee"], output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="bee", no_args_is_help=True)
def bee_group() -> None:
    """Group of sub-commands related to BEE."""


@bee_group.command(name="export", no_args_is_help=True)
@click.argument("config", type=click.Path(exists=True, readable=True))
def bee_export_command(config: str) -> None:
    """Generate BEE Images from YAML/JSON configuration.

    CONFIG is path to the YAML/JSON configuration

    The configuration template files could be generated by subcommand 'get-template'.
    """
    bee_export(config)


def bee_export(config: str) -> None:
    """Generate BEE Images from YAML/JSON configuration."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    schemas = BeeNxp.get_validation_schemas()
    check_config(config_data, schemas, search_paths=[config_dir])
    bee = BeeNxp.load_from_config(config_data, search_paths=[config_dir])

    output_folder = get_abs_path(config_data["output_folder"], config_dir)

    output_name = config_data.get("output_name")
    if not output_name:
        output_name = "encrypted.bin"

    bee_image = os.path.join(output_folder, output_name)
    write_file(bee.export_image(), bee_image, mode="wb")
    logger.info(f"Created BEE Image:\n{bee_image}")

    header_name = config_data.get("header_name")
    if not header_name:
        header_name = "bee_ehdr.bin"

    for idx, header in enumerate(bee.export_headers()):
        header_output = os.path.join(output_folder, f"{idx}_" + header_name)
        write_file(header, header_output, mode="wb")
        logger.info(f"Created BEE Header:\n{header_output}")

    click.echo(f"Success. BEE files have been created and stored into: {output_folder}")


@bee_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt1010",
    type=click.Choice(BeeNxp.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def bee_get_template_command(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    bee_get_template(chip_family, overwrite, output)


# pylint: disable=unused-argument    # preparation for the future
def bee_get_template(chip_family: str, overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(BeeNxp.generate_config_template(), output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="bootable-image", no_args_is_help=True)
def bootable_image_group() -> None:
    """Group of bootable image utilities."""


@bootable_image_group.command(name="merge", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Bootable Image YAML/JSON Configuration.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def bootable_image_merge_command(config: str, output: str) -> None:
    """Merge boot image blocks into one bootable image.

    The configuration template files could be generated by subcommand 'get-templates'.
    """
    bootable_image_merge(config, output)


def bootable_image_merge(config: str, output: str) -> None:
    """Merge boot image blocks into one bootable image."""
    config_data = load_configuration(config)
    config_dir = os.path.dirname(config)
    bimg_image = BootableImage.load_from_config(config_data, search_paths=[config_dir])
    bimg_image_info = bimg_image.image_info()

    write_file(bimg_image_info.export(), output, mode="wb")

    logger.info(f"Created Bootable Image:\n{bimg_image_info.info()}")
    logger.info(f"Created Bootable Image memory map:\n{bimg_image_info.draw()}")
    click.echo(f"Success. (Bootable Image: {output} created.)")


@bootable_image_group.command(name="parse", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt5xx",
    type=click.Choice(BootableImage.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-m",
    "--mem-type",
    default="flexspi_nor",
    type=click.Choice(["flexspi_nor", "flexspi_nand", "semc_nand"], case_sensitive=False),
    required=True,
    help="Select the chip used memory type.",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to binary Bootable image to parse.",
)
@click.argument("parsed-data-path", type=click.Path(resolve_path=True))
def bootable_image_parse_command(
    chip_family: str, mem_type: str, binary: str, parsed_data_path: str
) -> None:
    """Parse Bootable Image into YAML configuration and binary images."""
    bootable_image_parse(chip_family, mem_type, binary, parsed_data_path)


def bootable_image_parse(
    chip_family: str, mem_type: str, binary: str, parsed_data_path: str
) -> None:
    """Parse Bootable Image into YAML configuration and binary images."""
    bimg_image = get_bimg_class(chip_family)(family=chip_family, mem_type=mem_type)
    bimg_image.parse(load_binary(binary))
    bimg_image_info = bimg_image.image_info()
    logger.info(f"Parsed Bootable image memory map: {bimg_image_info.draw()}")
    bimg_image.store_config(parsed_data_path)
    click.echo(
        f"Success. (Bootable Image: {binary} has been parsed and stored into {parsed_data_path} .)"
    )


@bootable_image_group.command(name="get-templates", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt595",
    type=click.Choice(BootableImage.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output_folder", type=click.Path(resolve_path=True))
def bootable_image_get_templates_command(
    chip_family: str, overwrite: bool, output_folder: str
) -> None:
    """Create template of configurations in YAML format from all memory types.

    The template files folder name is specified as argument of this command.
    """
    bootable_image_get_templates(chip_family, overwrite, output_folder)


def bootable_image_get_templates(chip_family: str, overwrite: bool, output_folder: str) -> None:
    """Create template of configurations in YAML format from all memory types."""
    mem_types = BootableImage.get_supported_memory_types(chip_family)
    for mem_type in mem_types:
        output = os.path.join(output_folder, f"bootimg_{chip_family}_{mem_type}.yaml")
        if not os.path.isfile(output) or overwrite:
            click.echo(f"Creating {output} template file.")
            write_file(BootableImage.generate_config_template(chip_family, mem_type), output)
        else:
            click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="hab", no_args_is_help=True)
def hab_group() -> None:  # pylint: disable=unused-argument
    """Group of sub-commands related to HAB container."""


@hab_group.command(name="export", no_args_is_help=True)
@click.option(
    "-c",
    "--command",
    type=click.Path(exists=True),
    required=True,
    help="BD configuration file to produce HAB container",
)
@click.option("-o", "--output", type=click.Path(resolve_path=True), help="Output file path.")
@click.argument("external", type=click.Path(), nargs=-1)
def hab_export_command(
    command: str,
    output: str,
    external: List[str],
) -> None:
    """Generate HAB container from configuration.

    EXTERNAL is a space separated list of external binary files defined in BD file
    """
    image = hab_export(command, external)
    write_file(image, output, mode="wb")
    click.echo(f"Success. (HAB container: {output} created.)")


def hab_export(command: str, external: List[str]) -> bytes:
    """Generate HAB container from configuration."""
    try:
        parser = bd_parser.BDParser()

        bd_file_content = load_text(command)
        bd_data = parser.parse(text=bd_file_content, extern=external)
        if bd_data is None:
            raise SPSDKError("Invalid bd file, secure binary file generation terminated")
        bd_config = ImageConfig.parse(bd_data)
        hab_container = HabContainer.load(
            bd_config,
            search_paths=[os.path.dirname(command)],
            timestamp=bd_config.options.signature_timestamp,
        )
        data = hab_container.export()
        return data
    except SPSDKError as exc:
        raise SPSDKAppError(f"The HAB container generation failed: ({str(exc)}).") from exc


@hab_group.command(name="parse", no_args_is_help=True)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary HAB image to parse.",
)
@click.argument("output_folder", type=click.Path(resolve_path=True))
def hab_parse_command(
    binary: str,
    output_folder: str,
) -> None:
    """Parse HAB contianer into individual segments."""
    file_bin = load_binary(binary)
    created_files = hab_parse(file_bin, output_folder)
    for file_path in created_files:
        click.echo(f"File has been created: {file_path}")
    click.echo(f"Success. (HAB container parsed into: {output_folder}.)")


def hab_parse(binary: bytes, output: str) -> List[str]:
    """Generate HAB container from configuration."""
    hab_container = HabContainer.parse(binary)
    created_files = []
    for sub_image in hab_container.hab_image.sub_images:
        sub_image_data = sub_image.export()
        sub_image_path = os.path.join(output, f"{sub_image.name.lower()}.bin")
        write_file(sub_image_data, sub_image_path, mode="wb")
        created_files.append(sub_image_path)
    return created_files


@bootable_image_group.group(name="fcb", no_args_is_help=True)
def fcb() -> None:  # pylint: disable=unused-argument
    """FCB (Flash Configuration Block) utilities."""


@fcb.command(name="export", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="FCB YAML/JSON Configuration.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def fcb_export_command(config: str, output: str) -> None:
    """Export FCB Image from YAML/JSON configuration.

    The configuration template files could be generated by subcommand 'get-templates'.
    """
    fcb_export(config, output)


def fcb_export(config: str, output: str) -> None:
    """Export FCB Image from YAML/JSON configuration."""
    config_data = load_configuration(config)
    check_config(config_data, FCB.get_validation_schemas_family())
    chip_family = config_data["family"]
    mem_type = config_data["type"]
    revision = config_data.get("revision", "latest")
    schemas = FCB.get_validation_schemas(chip_family, mem_type, revision)
    check_config(config_data, schemas, search_paths=[os.path.dirname(config)])
    fcb_image = FCB.load_from_config(config_data)
    fcb_data = fcb_image.export()
    write_file(fcb_data, output, mode="wb")

    logger.info(f"Created FCB Image:\n{fcb_image.registers.image_info().info()}")
    logger.info(f"Created FCB Image memory map:\n{fcb_image.registers.image_info().draw()}")
    click.echo(f"Success. (FCB: {output} created.)")


@fcb.command(name="parse", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt5xx",
    type=click.Choice(FCB.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-m",
    "--mem-type",
    default="flexspi_nor",
    type=click.Choice(["flexspi_nor"], case_sensitive=False),
    required=True,
    help="Select the chip used memory type.",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary FCB image to parse.",
)
@click.argument("config-file", type=click.Path(resolve_path=True))
def fcb_parse_command(chip_family: str, mem_type: str, binary: str, config_file: str) -> None:
    """Parse FCB Image into YAML configuration."""
    fcb_parse(chip_family, mem_type, binary, config_file)


def fcb_parse(chip_family: str, mem_type: str, binary: str, config_file: str) -> None:
    """Parse FCB Image into YAML configuration."""
    fcb_image = FCB(family=chip_family, mem_type=mem_type)
    fcb_image.parse(load_binary(binary))

    logger.info(f"Parsed FCB image memory map: {fcb_image.registers.image_info().draw()}")
    config = fcb_image.create_config()
    write_file(config, config_file)
    click.echo(f"Success. (FCB: {binary} has been parsed and stored into {config_file} .)")


@fcb.command(name="get-templates", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt595",
    type=click.Choice(FCB.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output_folder", type=click.Path(resolve_path=True))
def fcb_get_templates_command(chip_family: str, overwrite: bool, output_folder: str) -> None:
    """Create template of configurations in YAML format for all memory types.

    The template files folder name is specified as argument of this command.
    """
    fcb_get_templates(chip_family, overwrite, output_folder)


def fcb_get_templates(chip_family: str, overwrite: bool, output_folder: str) -> None:
    """Create template of configurations in YAML format for all memory types."""
    mem_types = FCB.get_supported_memory_types(chip_family)
    for mem_type in mem_types:
        output = os.path.join(output_folder, f"fcb_{chip_family}_{mem_type}.yaml")
        if not os.path.isfile(output) or overwrite:
            click.echo(f"Creating {output} template file.")
            write_file(FCB.generate_config_template(chip_family, mem_type), output)
        else:
            click.echo(f"Skip creating {output}, this file already exists.")


@bootable_image_group.group(name="xmcd", no_args_is_help=True)
def xmcd() -> None:  # pylint: disable=unused-argument
    """XMCD (External Memory Configuration Data) utilities."""


@xmcd.command(name="export", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="XMCD YAML/JSON Configuration.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def xmcd_export_command(config: str, output: str) -> None:
    """Export XMCD Image from YAML/JSON configuration.

    The configuration template files could be generated by subcommand 'get-templates'.
    """
    xmcd_export(config, output)


def xmcd_export(config: str, output: str) -> None:
    """Export XMCD Image from YAML/JSON configuration."""
    config_data = load_configuration(config)
    xmcd_image = XMCD.load_from_config(config_data)
    xmcd_data = xmcd_image.export()
    write_file(xmcd_data, output, mode="wb")

    logger.info(f"Created XMCD :\n{xmcd_image.registers.image_info().info()}")
    logger.info(f"Created XMCD memory map:\n{xmcd_image.registers.image_info().draw()}")
    click.echo(f"Success. (XMCD: {output} created.)")


@xmcd.command(name="parse", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt117x",
    type=click.Choice(XMCD.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    required=True,
    help="Path to binary XMCD image to parse.",
)
@click.argument("config-file", type=click.Path(resolve_path=True))
def xmcd_parse_command(chip_family: str, binary: str, config_file: str) -> None:
    """Parse XMCD Image into YAML configuration."""
    xmcd_parse(chip_family, binary, config_file)


def xmcd_parse(chip_family: str, binary: str, config_file: str) -> None:
    """Parse XMCD Image into YAML configuration."""
    xmcd_image = XMCD(family=chip_family)
    xmcd_image.parse(load_binary(binary))

    logger.info(f"Parsed XMCD memory map: {xmcd_image.registers.image_info().draw()}")
    config = xmcd_image.create_config()
    write_file(config, config_file)
    click.echo(f"Success. (XMCD: {binary} has been parsed and stored into {config_file} .)")


@xmcd.command(name="get-templates", no_args_is_help=True)
@click.option(
    "-f",
    "--chip-family",
    default="rt117x",
    type=click.Choice(XMCD.get_supported_families(), case_sensitive=False),
    required=True,
    help="Select the chip family.",
)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output_folder", type=click.Path(resolve_path=True))
def xmcd_get_templates_command(chip_family: str, overwrite: bool, output_folder: str) -> None:
    """Create template of configurations in YAML format for all memory types.

    The template files folder name is specified as argument of this command.
    """
    xmcd_get_templates(chip_family, overwrite, output_folder)


def xmcd_get_templates(chip_family: str, overwrite: bool, output_folder: str) -> None:
    """Create template of configurations in YAML format for all memory types."""
    mem_types = XMCD.get_supported_memory_types(chip_family)
    for mem_type in mem_types:
        config_types = XMCD.get_supported_configuration_types(chip_family, mem_type)
        for config_type in config_types:
            output = os.path.join(
                output_folder, f"xmcd_{chip_family}_{mem_type}_{config_type}.yaml"
            )
            if not os.path.isfile(output) or overwrite:
                click.echo(f"Creating {output} template file.")
                write_file(
                    XMCD.generate_config_template(chip_family, mem_type, config_type), output
                )
            else:
                click.echo(f"Skip creating {output}, this file already exists.")


@main.group(name="utils", no_args_is_help=True)
def utils_group() -> None:
    """Group of utilities."""


@utils_group.group(name="binary-image", no_args_is_help=True)
def bin_image_group() -> None:
    """Binary Image utilities."""


@bin_image_group.command(name="create", no_args_is_help=True)
@click.option(
    "-s",
    "--size",
    type=INT(),
    required=True,
    help="Size of file to be created.",
)
@click.option(
    "-p",
    "--pattern",
    type=str,
    default="zeros",
    help="Pattern of created file ('zeros', 'ones', 'rand', 'inc' or any number value).",
)
@click.argument("output", type=click.Path(resolve_path=True))
def binary_create_command(size: int, pattern: str, output: str) -> None:
    """Create binary file with pattern.

    The helper utility to create simple binary file with pattern.
    """
    binary_create(size, pattern, output)


def binary_create(size: int, pattern: str, output: str) -> None:
    """Create binary file with pattern."""
    image = BinaryImage(name="", size=size, pattern=BinaryPattern(pattern))
    image.validate()
    data = image.export()

    write_file(data, output, mode="wb")

    logger.info(f"Created file:\n{image.info()}")
    logger.info(f"Created file:\n{image.draw()}")
    click.echo(f"Success. (Created binary file: {output} )")


@bin_image_group.command(name="merge", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to configuration file of merge operation.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def binary_merge_command(config: str, output: str) -> None:
    """Merge binary images together by description in YAML/JSON configuration.

    The configuration template files could be generated by subcommand 'utils bin-image get-template'.
    """
    binary_merge(config, output)


def binary_merge(config: str, output: str) -> None:
    """Merge binary images together by description in YAML/JSON configuration."""
    cfg = load_configuration(config)
    config_dir = os.path.dirname(config)
    check_config(cfg, BinaryImage.get_validation_schemas(), search_paths=[config_dir])
    image = BinaryImage.load_from_config(cfg, search_paths=[config_dir])
    try:
        image.validate()
    except SPSDKError as exc:
        click.echo(f"Image Validation fail:\n{image.draw()}")
        raise SPSDKError("Image validation failed") from exc
    data = image.export()

    write_file(data, output, mode="wb")

    logger.info(f"Merged Image:\n{image.info()}")
    logger.info(f"Merged Image:\n{image.draw()}")
    click.echo(f"Success. (Merged image: {output} created.)")


@bin_image_group.command(name="extract", no_args_is_help=True)
@click.option(
    "-b",
    "--binary",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to binary file to be used to extract chunk from.",
)
@click.option(
    "-a",
    "--address",
    type=str,
    required=True,
    help="Address of extracted chunk.",
)
@click.option(
    "-s",
    "--size",
    type=str,
    required=True,
    help="Size of extracted chunk. For '0' it extract rest of the file from given address.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def binary_extract_command(binary: str, address: str, size: str, output: str) -> None:
    """Extract chunk from binary file.

    \b
    OUTPUT - File name of stored chunk.
    """
    binary_extract(binary, address, size, output)


def binary_extract(binary: str, address: str, size: str, output: str) -> None:
    """Extract chunk from binary file."""
    bin_data = load_binary(binary)
    start = value_to_int(address)
    size_int = value_to_int(size)
    if not size_int:
        size_int = len(bin_data) - start
    end = start + size_int

    if end > len(bin_data):
        click.echo(f"The required binary chunk is out of [{binary}] file space.")
        return
    write_file(bin_data[start:end], output, mode="wb")
    click.echo(f"Success. (Extracted chunk: {output} created.)")


@bin_image_group.command(name="convert", no_args_is_help=True)
@click.option(
    "-i",
    "--input-file",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to BIN/HEX/S19/ELF file to be converted.",
)
@click.option(
    "-f",
    "--format",
    "output_format",
    type=click.Choice(["BIN", "HEX", "S19"], case_sensitive=False),
    required=True,
    help="Output format.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def binary_convert_command(input_file: str, output_format: str, output: str) -> None:
    """Convert input data file into selected format.

    \b
    OUTPUT - File name of converted file.
    """
    binary_convert(input_file, output_format, output)


def binary_convert(input_file: str, output_format: str, output: str) -> None:
    """Convert input data file into selected format."""
    image = BinaryImage.load_binary_image(input_file)
    logger.info(image.draw())
    image.save_binary_image(output, file_format=output_format)
    click.echo(f"Success. (Converted file: {output} created.)")


@bin_image_group.command(name="get-template", no_args_is_help=True)
@click.option(
    "-o",
    "--overwrite",
    default=False,
    type=bool,
    is_flag=True,
    help="Allow overwriting existing template file.",
)
@click.argument("output", type=click.Path(resolve_path=True))
def binary_get_template_command(overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format.

    The template file name is specified as argument of this command.
    """
    binary_get_template(overwrite, output)


def binary_get_template(overwrite: bool, output: str) -> None:
    """Create template of configuration in YAML format."""
    if not os.path.isfile(output) or overwrite:
        click.echo(f"Creating {output} template file.")
        write_file(BinaryImage.generate_config_template(), output)
    else:
        click.echo(f"Skip creating {output}, this file already exists.")


@bin_image_group.command(name="align", no_args_is_help=True)
@click.option(
    "-i",
    "--input",
    "input_file",
    type=click.Path(file_okay=True, dir_okay=False, resolve_path=True),
    required=True,
    help="Binary file name to be aligned.",
)
@click.option(
    "-o",
    "--output",
    "output_file",
    type=click.Path(resolve_path=True),
    help="Aligned file name. If not specified, input file will be updated.",
)
@click.option(
    "-a",
    "--alignment",
    default=1,
    type=INT(),
    help="Alignment size.",
)
@click.option(
    "-p",
    "--pattern",
    type=str,
    default="zeros",
    help="Pattern of used padding ('zeros', 'ones', 'rand', 'inc' or any number value).",
)
def binary_align_command(input_file: str, output_file: str, alignment: int, pattern: str) -> None:
    """Align binary file to provided alignment and padded with specified pattern."""
    binary_align(input_file, output_file, alignment, pattern)


def binary_align(
    input_file: str, output_file: Optional[str] = None, alignment: int = 1, pattern: str = "zeros"
) -> None:
    """Align binary file to provided alignment and padded with specified pattern.

    :param input_file: Input file name.
    :param output_file: Output file name. If not specified, input file will be updated.
    :param alignment: Size of alignment block.
    :param pattern: Padding pattern.

    :raises SPSDKError: Invalid input arguments.
    """
    if not input_file or not os.path.isfile(input_file):
        raise SPSDKError("Binary file alignment: Invalid input file")
    if alignment <= 0:
        raise SPSDKError("Binary file alignment: Alignment value must be 1 or greater.")
    if not output_file:
        output_file = input_file
    binary = align_block(data=load_binary(input_file), alignment=alignment, padding=pattern)
    write_file(binary, output_file, mode="wb")
    click.echo(
        f"The {input_file} file has been aligned to {len(binary)} ({hex(len(binary))}) bytes \
            and stored into {output_file}"
    )


@bin_image_group.command(name="pad", no_args_is_help=True)
@click.option(
    "-i",
    "--input",
    "input_file",
    type=click.Path(file_okay=True, dir_okay=False, resolve_path=True),
    required=True,
    help="Binary file name to be padded.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True),
    help="Padded file name. If not specified, input file will be updated.",
)
@click.option(
    "-s",
    "--size",
    required=True,
    type=INT(),
    help="Result file size.",
)
@click.option(
    "-p",
    "--pattern",
    type=str,
    default="zeros",
    help="Pattern of used padding ('zeros', 'ones', 'rand', 'inc' or any number value).",
)
def binary_pad_command(input_file: str, output: str, size: int, pattern: str) -> None:
    """Pad binary file to provided final size with specified pattern."""
    binary_pad(input_file, output, size, pattern)


def binary_pad(
    input_file: str, output: Optional[str] = None, size: int = 1, pattern: str = "zeros"
) -> None:
    """Pad binary file to provided final size with specified pattern.

    :param input_file: Input file name.
    :param output: Output file name. If not specified, input file will be updated.
    :param size: Final size of file.
    :param pattern: Padding pattern.

    :raises SPSDKError: Invalid input arguments.
    """
    if not input_file or not os.path.isfile(input_file):
        raise SPSDKError("Binary file alignment: Invalid input file")
    if not output:
        output = input_file
    binary = load_binary(input_file)
    if len(binary) < size:
        binary = align_block(data=binary, alignment=size, padding=pattern)
    write_file(binary, output, mode="wb")

    click.echo(
        f"The {input_file} file has been padded to {len(binary)} ({hex(len(binary))}) byte size\
            and stored into {output}"
    )


@utils_group.group(name="convert", no_args_is_help=True)
def convert() -> None:  # pylint: disable=unused-argument
    """Conversion utilities."""


@convert.command(name="hex2bin", no_args_is_help=True)
@click.option(
    "-i",
    "--input-file",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to text file with hexadecimal string to be converted to binary.",
)
@click.option(
    "-r",
    "--reverse",
    type=bool,
    is_flag=True,
    default=False,
    help="The resulting binary bytes will be stored in reverse order (for example SBKEK in elftosb requires that).",
)
@click.argument("output", type=click.Path(resolve_path=True))
def convert_hex2bin_command(input_file: str, reverse: bool, output: str) -> None:
    """Convert file with hexadecimal string into binary file with optional reverse order of stored bytes.

    \b
    OUTPUT - File name of converted file.
    """
    convert_hex2bin(input_file, reverse, output)


def convert_hex2bin(input_file: str, reverse: bool, output: str) -> None:
    """Convert file with hexadecimal string into binary file with optional reverse order of stored bytes."""
    try:
        value = bytearray.fromhex(load_text(input_file))
    except (SPSDKError, ValueError) as e:
        raise SPSDKAppError(f"Failed loading hexadecimal value from: {input_file}") from e
    else:
        if reverse:
            value.reverse()
        write_file(value, output, mode="wb")
        click.echo(f"Success. Converted file: {output}")


@convert.command(
    name="bin2hex",
    no_args_is_help=True,
    short_help="Convert binary file to hexadecimal text file.",
    help="Convert binary file into hexadecimal text file with optional reverse order of stored bytes.",
)
@click.option(
    "-i",
    "--input-file",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to binary file with to be converted to hexadecimal text file.",
)
@click.option(
    "-r",
    "--reverse",
    type=bool,
    is_flag=True,
    default=False,
    help="The result binary bytes will be stored in reverse order (for example SBKEK in elftosb this required).",
)
@click.argument("output", type=click.Path(resolve_path=True))
def convert_bin2hex_command(input_file: str, reverse: bool, output: str) -> None:
    """Convert binary file into hexadecimal text file with optional reverse order of stored bytes.

    \b
    OUTPUT - File name of converted file.
    """
    convert_bin2hex(input_file, reverse, output)


def convert_bin2hex(input_file: str, reverse: bool, output: str) -> None:
    """Convert binary file into hexadecimal text file with optional reverse order of stored bytes."""
    value = bytearray(load_binary(input_file))
    if reverse:
        value.reverse()
    write_file(value.hex(), output, mode="w")
    click.echo(f"Success. Converted file: {output}")


@convert.command(name="bin2carr", no_args_is_help=True)
@click.option(
    "-i",
    "--input-file",
    type=click.Path(exists=True, readable=True),
    required=True,
    help="Path to binary file to be converted to C array.",
)
@click.option(
    "-n",
    "--name",
    type=str,
    help="The output C array name, if not specified the name of input file will be used.",
)
@click.option(
    "-t",
    "--type",
    "output_type",
    type=click.Choice(["uint8_t", "uint16_t", "uint32_t", "uint64_t"]),
    default="uint8_t",
    help="The output C array type.",
)
@click.option(
    "-e",
    "--endian",
    type=click.Choice(["big", "little"]),
    default="big",
    help="The binary input file endian.",
)
@click.option(
    "-p",
    "--padding",
    type=str,
    help="The padding value in case of non aligned binary image.",
)
@click.option(
    "-c",
    "--count-per-line",
    type=click.IntRange(1, 1024),
    default=8,
    help="The array items count per line.",
)
@click.option(
    "--tab",
    type=click.IntRange(0, 32),
    default=4,
    help="The tabulator size for new line.",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(resolve_path=True),
    help=(
        "Path to output text file with created C array, "
        "if not specified the output will be printed into standard output."
    ),
)
def convert_bin2carr(
    input_file: str,
    name: str,
    output_type: str,
    endian: str,
    padding: str,
    count_per_line: int,
    tab: int,
    output_file: str,
) -> None:
    """Convert binary file into C array string.

    Default output is byte representation, but it could be
    converted to 16/32 or 64 bit unsigned types with specified endianness.
    """
    raw_binary = load_binary(input_file)
    width = {"uint8_t": 1, "uint16_t": 2, "uint32_t": 4, "uint64_t": 8}[output_type]

    if len(raw_binary) != align(len(raw_binary), width) and padding is None:
        raise SPSDKAppError("Unaligned binary image, if still to be used, define '-p' padding.")

    binary = align_block(
        data=raw_binary, alignment=width, padding=BinaryPattern(padding) if padding else None
    )
    rest_bytes = len(binary)
    count = rest_bytes // width
    name = name or os.path.splitext(os.path.basename(input_file))[0]
    index = 0
    ret = f"const {output_type} {name}[{count}] = {{\n"
    while rest_bytes:
        line_cnt = min(count_per_line, rest_bytes // width)
        comment = f"// 0x{index:09_X}"
        ret += " " * tab
        for _ in range(line_cnt):
            data = bytearray(binary[index : index + width])
            if endian.lower() == "little":
                data.reverse()
            ret += "0x" + data.hex() + ", "
            index += width
            rest_bytes -= width
        ret += f" {comment}\n"
    ret += "};\n"

    if output_file:
        write_file(ret, output_file)
        click.echo(f"Success. Created C file: {output_file}")
    else:
        click.echo(ret)


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Loading methods for keys/certificates/CSR."""

from typing import Iterable, List, Optional

from spsdk import SPSDKError
from spsdk.crypto import (
    Certificate,
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
    Encoding,
    PrivateKey,
    PublicKey,
    RSAPrivateKey,
    RSAPublicKey,
    _PublicKeyTuple,
    default_backend,
    load_der_private_key,
    load_der_public_key,
    load_der_x509_certificate,
    load_pem_private_key,
    load_pem_public_key,
    load_pem_x509_certificate,
)
from spsdk.utils.misc import load_binary


def load_private_key_from_data(
    data: bytes, password: Optional[bytes] = None, encoding: Optional[Encoding] = None
) -> PrivateKey:
    """Load private key from bytes.

    :param data: data of private key loaded from file
    :param password: password for key
    :param encoding: encoding type of key
    :return: RSA private key
    :raises SPSDKError: Unsupported private key to load
    """
    real_encoding = encoding or _get_encoding_type(data)

    try:
        private_key = {
            Encoding.PEM: load_pem_private_key,
            Encoding.DER: load_der_private_key,
        }[
            real_encoding
        ](data, password, default_backend())
        assert isinstance(private_key, (RSAPrivateKey, EllipticCurvePrivateKey))
        return private_key
    except ValueError as exc:
        raise SPSDKError(f"Cannot load private key: ({str(exc)})") from exc


def load_private_key(
    file_path: str, password: Optional[bytes] = None, encoding: Optional[Encoding] = None
) -> PrivateKey:
    """Load private key from file.

    :param file_path: path to file, where private key is stored
    :param password: password for key
    :param encoding: encoding type of key
    :return: RSA/ECC private key
    """
    data = load_binary(file_path)
    return load_private_key_from_data(data, password, encoding)


def load_public_key_from_data(data: bytes, encoding: Optional[Encoding] = None) -> PublicKey:
    """Load the public key from bytes.

    :param data: data of public key loaded from file
    :param encoding: encoding type of key
    :return: RSA public key
    :raises SPSDKError: Unsupported public key to load
    """
    real_encoding = encoding or _get_encoding_type(data)

    try:
        public_key = {
            Encoding.PEM: load_pem_public_key,
            Encoding.DER: load_der_public_key,
        }[
            real_encoding
        ](data, default_backend())
        assert isinstance(public_key, (RSAPublicKey, EllipticCurvePublicKey))
        return public_key
    except ValueError as exc:
        raise SPSDKError(f"Cannot load public key: ({str(exc)})") from exc


def load_public_key(file_path: str, encoding: Optional[Encoding] = None) -> PublicKey:
    """Load the public key from file.

    :param file_path: path to file, where public key is stored
    :param encoding: encoding type of key
    :return: RSA public key
    """
    data = load_binary(file_path)
    return load_public_key_from_data(data, encoding)


def load_certificate_from_data(data: bytes, encoding: Optional[Encoding] = None) -> Certificate:
    """Load the certificate from bytes.

    :param data: data with certificate loaded from file
    :param encoding: type of encoding
    :return: Certificate (from cryptography library)
    :raises SPSDKError: Unsupported certificate to load
    """
    real_encoding = encoding or _get_encoding_type(data)
    try:
        return {
            Encoding.PEM: load_pem_x509_certificate,
            Encoding.DER: load_der_x509_certificate,
        }[
            real_encoding
        ](data, default_backend())
    except ValueError as exc:
        raise SPSDKError(f"Cannot load certificate: ({str(exc)})") from exc


def load_certificate(file_path: str, encoding: Optional[Encoding] = None) -> Certificate:
    """Load the certificate from file.

    :param file_path: path to file, where certificate is stored
    :param encoding: type of encoding
    :return: Certificate (from cryptography library)
    """
    data = load_binary(file_path)
    return load_certificate_from_data(data, encoding)


def load_certificate_as_bytes(file_path: str) -> bytes:
    """Load certificate from file in PEM/DER format.

    Converts the certificate into DER format and serializes it into bytes.

    :param file_path: path to certificate file.
    :return: certificate in der format serialized into bytes.
    """
    return load_certificate(file_path).public_bytes(encoding=Encoding.DER)


def _get_encoding_type(data: bytes) -> Encoding:
    """Get the encoding type out of given item from the data.

    :param file: name of file, where item is stored
    :return: encoding type (Encoding.PEM, Encoding.DER)
    """
    encoding = Encoding.PEM
    try:
        decoded = data.decode("utf-8")
    except UnicodeDecodeError:
        encoding = Encoding.DER
    else:
        if decoded.find("----") == -1:
            encoding = Encoding.DER
    return encoding


def extract_public_key_from_data(object_data: bytes, password: Optional[str] = None) -> PublicKey:
    """Extract any kind of public key from a data that contains Certificate, Private Key or Public Key.

    :raises SPSDKError: Raised when file can not be loaded
    :return: private key of any type
    """
    try:
        cert_candidate = load_certificate_from_data(object_data)
        public_key = cert_candidate.public_key()
        assert isinstance(public_key, _PublicKeyTuple)
        return public_key
    except SPSDKError:
        pass

    try:
        private_candidate = load_private_key_from_data(
            object_data, password.encode() if password else None
        )
        return private_candidate.public_key()
    except SPSDKError:
        pass

    try:
        public_candidate = load_public_key_from_data(object_data)
        return public_candidate
    except SPSDKError as exc:
        raise SPSDKError("Unable to load secret data.") from exc


def extract_public_key(
    file_path: str, password: Optional[str] = None, search_paths: Optional[List[str]] = None
) -> PublicKey:
    """Extract any kind of public key from a file that contains Certificate, Private Key or Public Key.

    :param file_path: File path to public key file.
    :param password: Optional password for encrypted Private file source.
    :param search_paths: List of paths where to search for the file, defaults to None
    :raises SPSDKError: Raised when file can not be loaded
    :return: Public key of any type
    """
    try:
        object_data = load_binary(file_path, search_paths=search_paths)
        return extract_public_key_from_data(object_data, password)
    except SPSDKError as exc:
        raise SPSDKError(f"Unable to load secret file '{file_path}'.") from exc


def extract_public_keys(
    secret_files: Iterable[str],
    password: Optional[str] = None,
    search_paths: Optional[List[str]] = None,
) -> List[PublicKey]:
    """Extract any kind of public key from files that contain Certificate, Private Key or Public Key.

    :param secret_files: List of file paths to public key files.
    :param password: Optional password for encrypted Private file source.
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: List of public keys of any type
    """
    return [
        extract_public_key(file_path=source, password=password, search_paths=search_paths)
        for source in secret_files
    ]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Utilities used by adapters."""

import logging
from typing import List, Optional, Set

from spsdk.crypto.certificate_management import X509NameConfig
from spsdk.mboot.interfaces.usb import RawHid, scan_usb
from spsdk.utils.misc import Timeout

from .. import SPSDKTpError
from .tptarget_blhost import TpTargetBlHost

logger = logging.getLogger(__name__)


def sanitize_common_name(name_config: X509NameConfig) -> None:
    """Adjust the COMMON_NAME for TrustProvisioning purposes.

    Base common name will be AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA-BB
    AA will be eventually replaced by UUID
    BB will be the certificate index (0-3)
    If the common name already contains some string, it will be used as a prefix
    """
    if isinstance(name_config, dict):
        subject_cn = name_config.get("COMMON_NAME") or ""
        assert isinstance(subject_cn, str)
        name_config["COMMON_NAME"] = subject_cn + 16 * "AA" + "-" + "BB"

    if isinstance(name_config, list):

        def find_item_index(config: List, item_key: str) -> int:
            for i, item in enumerate(config):
                assert isinstance(item, dict)
                if item_key in item:
                    return i
            return -1

        subject_cn_idx = find_item_index(name_config, "COMMON_NAME")
        subject_cn = name_config[subject_cn_idx].get("COMMON_NAME") or ""
        subject_cn = subject_cn + 16 * "AA" + "-" + "BB"
        if subject_cn_idx == -1:
            name_config.append({"COMMON_NAME": subject_cn})
        else:
            name_config[subject_cn_idx] = {"COMMON_NAME": subject_cn}


def get_current_usb_paths() -> Set[bytes]:
    """Get paths to all NXP USB devices."""
    return {device.path for device in scan_usb()}


def detect_new_usb_path(
    initial_set: Optional[Set[bytes]] = None, timeout: int = 1000
) -> Optional[bytes]:
    """Return USB path to newly found NXP USB device.

    :param initial_set: Initial set of USB device paths, defaults to None
    :param timeout: Timeout in milliseconds for the device detection
    :raises SPSDKTpError: Unable to determine single USB device change in time
    :raises SPSDKTpError: Multiple USB devices detected at once
    :return: USB path to newly detected device, None in case no changes were detected
    """
    loc_timeout = Timeout(timeout=timeout, units="ms")
    previous_set = initial_set or set()
    while not loc_timeout.overflow():
        new_set = get_current_usb_paths()
        addition = new_set.difference(previous_set)
        logger.info(f"Additions: {addition}")
        previous_set = new_set
        if len(addition) > 1:
            raise SPSDKTpError("Multiple new usb devices detected at once!")
        if len(addition) == 1:
            return addition.pop()

    # when timeout passes and the USB paths set stays the same
    # this happens mostly on Windows under higher CPU load
    if initial_set == previous_set:
        logger.info("No changes were detected")
        return None

    raise SPSDKTpError("USB device detected malfunctioned")


def update_usb_path(tptarget: TpTargetBlHost, new_usb_path: Optional[bytes]) -> None:
    """Update USB path in TP target's MBoot USB."""
    if not isinstance(tptarget.mboot._device, RawHid):
        return
    if new_usb_path is None:
        return
    tptarget.mboot._device.path = new_usb_path
